import numpy as numpy
import inspect

from .IterativeFitter import IterativeFitter
from .kernels.Kernel import Kernel
from .kernels.Biweight import Biweight

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2008 - 2014 Do Kester, SRON (Java code)
#  *    2017        Do Kester


class RobustShell( IterativeFitter ):
    """
    RobustShell tries to make a fit more robust in the presence of outliers.

    It is a shell around another fitter. Technically it is in itself a
    a Fitter, but with limited possiblities.

    A RobustShell tries to make a fit more robust in the presence of outliers.
    It does it by manipulating the weights: outliers are
    downweighted. "Normal" points keep their weights, more or less.

    Apart from methods specific to the robustification, RobustShell has a fit method
    and little else from the Fitter family. Methods to get the :math:`\chi^2`,
    the covariance matrix, the evidence, the noise scale etc. should be taken from
    the embedded fitter.

    The theory behind robust fitting can be found in Wikipedia: Robust Statistics,
    and the references therein.

    To determine which points are "normal" and which are "outliers" we
    need a previous fit. The difference to this earlier fit determines
    the normalcy. Its amount is used to adjust the weights.

    The fact that we need a previous iteration makes robust estimation a
    iterative procedure.
    Typically the procedure should stabilize in half a dozen steps maximally.

    There are several schemes to adjust the weights. But before we go into that
    we need two values in each scheme.
    Firstly the amount of noise present in the data. By default the noise
    is taken from the previous fit via Fitter.scale.
    The second value needed is the size of the influence domain in terms
    of the noise scale.

    For all schemes the deviant is calculated as the difference
    between data and fit divided by noisescale and domainsize:

    d = ( data - fit ) / ( noisescale * domainsize )

    The domainsize is defined such that deviants upto 3 times the noisescale fall
    within the fwhm.

    A number of weighting schemes are provided.

    With bound support and smooth edges:
    .. Kernel       Name        domain      comment
    .. Biweight     Tukey         5.54      Default kernel
    .. CosSquare                  6.00
    .. Tricube                    5.08
    .. Triweight                  6.60

    With bound support and hard edges:
    .. Uniform      Clip          3.00      Ignore all outside 3 sigma
    .. Cosine                     4.50
    .. Triangle                   6.00
    .. Parabola                   4.50

    with unbound support:
    .. Huber        Median        1.50      Inside domain mean; outside domain median
    .. Gauss                      2.12
    .. Lorentz                    3.00

    Other schemes can be written by making another Kernel or writing a function
        wgts = func( d )
    where d is the deviant as above.


    Notes
    -----
    Robust fitting is even more dangerous than ordinary fitting.
    *Never trust what you get without thorough checking.*

    Example
    -------
    >>> model = PolynomialModel( 1 )                # some model
    >>> x = numpy.arange( 100, dtype=float ) / 100  # some x values
    >>> y = numpy.arange( 100, dtype=float ) / 4    # digitization noise
    >>> y[1,11,35,67] += 10                         # create outliers
    >>> ftr = Fitter( x, model )                    # a fitter, a model and a x
    >>> rob = RobustShell( ftr, domain=7 )          # robust fitter using someFtr
    >>> par = rob.fit( y )                          # solution
    >>> print( rob )                                #
    >>> print( rob.weights )                        # print final weights
    >>> print( ftr.chisq )                          # get from someFtr

    Author       Do Kester.
    =============================================================================

    """
    def __init__( self, fitter, kernel=Biweight, domain=None, onesided=None, **kwargs ):
        """
        Create a new class, providing the fitter to be used.

        Parameters
        ----------
        fitter : BaseFitter
             to be used
        kernel : Kernel or callable
            All Kernels have a method `result( d )` which is applied to the deviants.
            where d = ( data - model ) / ( domain * scale )
            If kernel is a callable method it is assumed to be a similar result mathod.
        domain : None or float
            Width of the kernel.
            None : automatic calculation of domain according to table in class doc.
            float : overrides autocalculation.
        onesided : None or "positive" or "p" or "negative" or "n"
            None : apply robust weights to positive and negative residuals
            "positive" : apply robust weights to positive residuals only
            "negative" : apply robust weights to negative residuals only

        """
        super( RobustShell, self ).__init__( fitter.xdata, fitter.model, **kwargs )
        self.fitter = fitter
        self.kernelfunc = self.setKernel( kernel )
        if domain is None :
            domain = 6.0 / self.kernel.fwhm
        self.domain = domain
        self.onesided = self.setOneSided( onesided )

    def setKernel( self, kernel ) :
        """
        Set the robust kernel to be used.

        Parameters
        ----------
        kernel : Kernel or callable
            All Kernels have a method `result( d )` which is applied to the deviants.
            where d = ( data - model ) / ( domain * scale )
            If kernel is a callable method it is assumed to be a similar result mathod.

        Raises
        ------
        ValueError when kernel is not recognized.

        """
        if inspect.isclass( kernel ) :
            kernel = kernel()
        if isinstance( kernel, Kernel ) :
            self.kernel = kernel
            kernelfunc = self.kernel.result
        elif callable( kernel ) :
            self.kernel = None
            kernelfunc = kernel
        else :
            raise ValueError( "Could not interpret the value for kernel: %s" % str( kernel ) )

        return kernelfunc

    def setOneSided( self, onesided ) :
        """
        set self.onesided to either 0 or +1 or -1.

        Parameters
        ----------
        onesided : None or "positive" or "negative"
            None : apply robust weights to positive and negative residuals
            "positive" : apply robust weights to positive residuals only
            "negative" : apply robust weights to negative residuals only

        Raises
        ------
        ValueError when onesided could not be interpreted.

        """
        if onesided is None :
            return 0
        elif isinstance( onesided, str ) :
            if onesided[0] is "p" or onesided[0] is "P" :
                return +1
            elif onesided[0] is "n" or onesided[0] is "N" :
                return -1
            else :
                raise ValueError( "Unknown string for onesided: %s" % onesided )
        else :
            raise ValueError( "Could not interpret the value for onesided: %s" % str( onesided ) )


    def fit( self, data, weights=None, kernel=None, domain=None, onesided=None, **kwargs ) :
        """
        Perform a robustification step.

        Parameters
        ----------
        data : array_like
            the data as they go into a fitter
        kwargs : dict
            keyword args to be passed to fitter.fit()
        """

        kernelfunc = self.kernelfunc if kernel is None else self.setKernel( kernel )
        onesided = self.onesided if onesided is None else self.setOneSided( onesided )
        domain = self.domain if domain is None else domain

        self.iter = 0

        param = self.fitter.fit( data, weights, **kwargs )
        self.npfit = self.fitter.npfit
        chi = self.fitter.chisq

        while self.iter < self.maxIter :

            residuals = data - self.model.result( self.xdata, param )
            scale = self.fitter.scale
            residuals /= ( scale * domain )
            robwgt = kernelfunc( residuals )
            robwgt = self.getOneSidedWeights( robwgt, residuals, onesided )

            self.weights = robwgt
            if weights is not None :
                self.weights *= weights
            param = self.fitter.fit( data, self.weights, **kwargs )

            trychi = self.fitter.chisq
            tol = self.tolerance if trychi < 1 else self.tolerance * trychi

            if abs( chi - trychi ) < tol :
                self.report( self.verbose, param, trychi, more=scale, force=True )
                break

            self.report( self.verbose, param, trychi, more=scale, force=(self.verbose>=3) )
            chi = trychi
            self.iter += 1

        self.hessian = self.fitter.getHessian()
        self.chisq = trychi

        return param

    def getOneSidedWeights( self, wgt, res, onesided ) :
        if onesided == 0 :
            return wgt
        if onesided == +1 :
            return numpy.where( res < 0, 1.0, wgt )
        if onesided == -1 :
            return numpy.where( res > 0, 1.0, wgt )
        return wgt

    def __getattr__( self, name ) :
        return super( RobustShell, self ).__getattr__( name )

    def __str__( self ):
        """
        Return the name and weight of the fitter.
        """
        name = "RobustShell( " + str( self.fitter ) + " ) using a "
        if self.onesided == 0 :
            return name + self.kernel.name()
        if self.onesided == +1 :
            return name + " +1 sided " + self.kernel.name( )
        if self.onesided == -1 :
            return name + " -1 sided " + self.kernel.name( )
        return name + self.kernel.name()




import numpy as numpy
from astropy import units
import math
from . import Tools
from .NonLinearModel import NonLinearModel
from .kernels.Kernel import Kernel
from .kernels.Gauss import Gauss

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2017 Do Kester

class Kernel2dModel( NonLinearModel ):
    """
    Two dimensional Kernel Model.

    The Kernel2dModel is defined as
    .. math::
        f( x:p ) = p_0 * K( r )

    where K( r ) is a selectable kernel function and r is the distance to the center.
    .. math::
        r = sqrt( u^2 + v^2 ).

    There are 3 options for u and v

    1. CIRCULAR has 4 parameters
        Circular shape with only one width.
        :math: u = ( x - p_1 ) / p_3
        :math: v = ( x - p_2 ) / p_3
    2. ELLIPTIC has 5 parameters
        Elliptic shape aligned along the axes; 2 widths.
        :math: u = ( x - p_1 ) / p_3
        :math: u = ( x - p_2 ) / p_4
    3. ROTATED has 6 parameters
        Rotated elliptical shape with 2 width and a rotational angle.
        :math: u = ( ( x - p_1 )*cos( p_5 ) - ( y - p_2 )*sin( p_5) ) / p_3
        :math: v = ( ( x - p_1 )*sin( p_5 ) + ( y - p_2 )*cos( p_5) ) / p_4

    The "center" parameters ( 1&2 ) and the "angle" parameter ( 5 ) are initilized as 0.
    The rotational angle is measured counterclockwise from the x-axis.
    The "width" parameters ( 3&4 ) are initialized as 1.0, except for the ROTATED case;
    then they need to be different ( 2.0 and 0.5 resp. ).
    Otherwise the model parameter "angle" is degenerate.
    The "amplitude" parameter is set to 1.0.

    Several kernel functions, K( x ) are defined in the directory fit/kernels.

    Beware: These models are unaware of anything outside their range.

    Author:      Do Kester

    Example
    -------
    >>> model = Kernel2dModel( )
    >>> model.setKernel( kernel=Lorentz )                        # circular Lorentz model.
    >>> model = Kernel2dModel( shape=Kernel2dModel.ROTATED )
    >>> model.setKernel( Gauss )                                 # equivalent to Gauss2DRotModel.

    Category:    mathematics/Fitting

    """
    SHAPENAMES = ["Circular", "Circular", "Elliptic", "Rotated"]

    CIRCULAR = 1
    ELLIPTIC = 2
    ROTATED = 3
    NPAR = [4,4,5,6]

    def __init__( self, kernel=Gauss(), shape=1, copy=None, **kwargs ):
        """
        Kernel Model.

        Default model: Gauss with Circular shape.

        Parameters
        ----------
        kernel : Kernel
            the kernel to be usel
        shape : 1 | 2 | 3
            shapes, resp.: circular elliptic, rotated
        copy : Kernel2dModel
            to be copied
        fixed : dictionary of {int:float}
            int     list if parameters to fix permanently. Default None.
            float   list of values for the fixed parameters.
            Attribute fixed can only be set in the constructor.

        """
        super( Kernel2dModel, self ).__init__( self.NPAR[shape], ndim=2, copy=copy, **kwargs )

        if copy is not None :
            self.kernel = copy.kernel
            self.shape = copy.shape
            self.shape2d = copy.shape2d
        else :
            self.setKernelShape( kernel, shape )

    def copy( self ):
        """ Copy method.  """
        return Kernel2dModel( copy=self, shape=self.shape )

    def __setattr__( self, name, value ) :
        if name == "kernel" and isinstance( value, Kernel ) :
            object.__setattr__( self, name, value )
        elif name == "shape2d" and isinstance( value, BaseShape2d ) :
            object.__setattr__( self, name, value )
        elif name == "shape" :
            object.__setattr__( self, name, value )
        else :
            super( Kernel2dModel, self ).__setattr__( name, value )

    def setKernelShape( self, kernel, shape ) :
        self.kernel = kernel
        amp = 2.0 / ( self.kernel.integral * math.pi )

        self.shape = shape
        self.npbase = self.NPAR[shape]
        if self.fixed is not None : self.npbase -= len( self.fixed )

        parNames = ["amplitude", "x-center", "y-center"]
        if shape == 1 :
            pars = Tools.toArray( [amp, 0.0, 0.0, 1.0] )
            self.parameters = self.select( pars )
            self.posIndex = [3]
            self.nonZero = [3]
            parNames += ["width"]
            self.shape2d = Circle( self.npbase, self.kernel )
        elif shape == 2 :
            pars = Tools.toArray( [amp, 0.0, 0.0, 1.0, 1.0] )
            self.parameters = self.select( pars )
            self.posIndex = [3,4]
            self.nonZero = [3,4]
            parNames += ["x-width", "y-width"]
            self.shape2d = Ellipse( self.npbase, self.kernel )
        elif shape == 3 :
            pars = Tools.toArray( [amp, 0.0, 0.0, 2.0, 0.5, 0.0] )
            self.parameters = self.select( pars )
            self.posIndex = [3,4]
            self.nonZero = [3,4]
            parNames += ["x-width", "y-width", "angle"]
            self.shape2d = Rotated( self.npbase, self.kernel )
        self.parNames = self.selectNames( parNames )

    def baseName( self ):
        """ Returns a string representation of the model.  """
        return "2-dim " + self.SHAPENAMES[self.shape] + "-" + self.kernel.name( )

    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the result
        params : array_like
            values for the parameters

        """
        return self.shape2d.result( xdata, params )

    def baseDerivative( self, xdata, params ):
        """
        Returns df/dx of the model function.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the result
        params : array_like
            values for the parameters

        """
        return self.shape2d.derivative( xdata, params )

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the xdata value.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the partials
        params : array_like
            parameters to the model (ignored in LinearModels)
        parlist : array_like
            list of indices of active parameters

        """
        return self.shape2d.partial( xdata, params, parlist=parlist )


    def baseParameterUnit( self, k ):
        """
        Return the unit of a parameter.
        Parameters
        ----------
        k : int
            the kth parameter.

        """
        if k == 0:
            return self.yUnit
        if k == 5:
            return units.Unit( units.si.rad )
        return self.xUnit[(k-1)%2]

    def isBound( self ):
        return self.kernel.isBound( )

class BaseShape2d( object ):
    def __init__( self, npbase, kernel ) :
        self.npbase = npbase
        self.kernel = kernel

    def reslt( self, xdata, params, kp ):
        x = ( xdata[:,0] - params[1] ) / params[3]
        y = ( xdata[:,1] - params[2] ) / params[kp]
        return params[0] * self.kernel.resultsq( x * x + y * y )

    def deriv( self, xdata, params, kp ) :
        x = ( xdata[:,0] - params[1] ) / params[3]
        y = ( xdata[:,1] - params[2] ) / params[kp]
        df = [params[0] * self.kernel.partial( x ) / params[3],
              params[0] * self.kernel.partial( y ) / params[kp]]
        return df


class Circle( BaseShape2d ):
    def __init__( self, npbase, kernel ) :
        super( Circle, self ).__init__( npbase, kernel )

    def result( self, xdata, params ):
        return self.reslt( xdata, params, 3 )

    def derivative( self, xdata, params ):
        return self.deriv( xdata, params, 3 )

    def partial( self, xdata, params, parlist=None ):
        np = self.npbase if parlist is None else len( parlist )
        partial = numpy.ndarray( ( Tools.length( xdata[:,0] ), np ) )
        u = ( xdata[:,0] - params[1] ) / params[3]
        v = ( xdata[:,1] - params[2] ) / params[3]
        r2 = u * u + v * v
        r = numpy.sqrt( r2 )
        p0dkdr = params[0] * self.kernel.partial( r )

        parts = { 0 : ( lambda: self.kernel.resultsq( r2 ) ),
                  1 : ( lambda: -p0dkdr * u / ( r * params[3] ) ),
                  2 : ( lambda: -p0dkdr * v / ( r * params[3] ) ),
                  3 : ( lambda: -p0dkdr * r / params[3] ) }

        if parlist is None :
            parlist = range( np )

        for k,kp in enumerate( parlist ) :
            partial[:,k] = parts[kp]()

        return partial

class Ellipse( BaseShape2d ):
    def __init__( self, npbase, kernel ) :
        super( Ellipse, self ).__init__( npbase, kernel )

    def result( self, xdata, params ):
        return self.reslt( xdata, params, 4 )

    def derivative( self, xdata, params ):
        return self.deriv( xdata, params, 4 )

    def partial( self, xdata, params, parlist=None ):
        np = self.npbase if parlist is None else len( parlist )
        partial = numpy.ndarray( ( Tools.length( xdata[:,0] ), np ) )
        u = ( xdata[:,0] - params[1] ) / params[3]
        v = ( xdata[:,1] - params[2] ) / params[4]
        r2 = u * u + v * v
        r = numpy.sqrt( r2 )
        p0dkdr = params[0] * self.kernel.partial( r )

        parts = { 0 : ( lambda: self.kernel.resultsq( r2 ) ),
                  1 : ( lambda: -p0dkdr * u / ( r * params[3] ) ),
                  2 : ( lambda: -p0dkdr * v / ( r * params[4] ) ),
                  3 : ( lambda: -p0dkdr * u * u / ( r * params[3] ) ),
                  4 : ( lambda: -p0dkdr * v * v / ( r * params[4] ) ) }


        if parlist is None :
            parlist = range( np )

        for k,kp in enumerate( parlist ) :
            partial[:,k] = parts[kp]()

        return partial

class Rotated( BaseShape2d ):
    def __init__( self, npbase, kernel ) :
        super( Rotated, self ).__init__( npbase, kernel )

    def result( self, xdata, params ):
        x = ( xdata[:,0] - params[1] )
        y = ( xdata[:,1] - params[2] )
        c = math.cos( params[5] )
        s = math.sin( params[5] )
        u = ( x * c - y * s ) / params[3]
        v = ( x * s + y * c ) / params[4]
        return params[0] * self.kernel.resultsq( u * u + v * v )

    def derivative( self, xdata, params ) :
        x = ( xdata[:,0] - params[1] )
        y = ( xdata[:,1] - params[2] )
        c = math.cos( params[5] )
        s = math.sin( params[5] )

        u = ( x * c - y * s ) / params[3]
        v = ( x * s + y * c ) / params[4]
        df = [params[0] * ( self.kernel.partial( u ) * c / params[3] +
                            self.kernel.partial( v ) * s / params[4] ),
              params[0] * ( self.kernel.partial( v ) * c / params[4] -
                            self.kernel.partial( u ) * s / params[3] )]
        return df

    def partial( self, xdata, params, parlist=None ):
        np = self.npbase if parlist is None else len( parlist )
        partial = numpy.ndarray( ( Tools.length( xdata[:,0] ), np ) )
        x = ( xdata[:,0] - params[1] )
        y = ( xdata[:,1] - params[2] )
        c = math.cos( params[5] )
        s = math.sin( params[5] )
        u = ( x * c - y * s ) / params[3]
        v = ( x * s + y * c ) / params[4]
        r2 = u * u + v * v
        r = numpy.sqrt( r2 )
        p0dkdr = params[0] * self.kernel.partial( r )
        drdu = -u / r
        drdv = -v / r

        parts = { 0 : ( lambda: self.kernel.resultsq( r2 ) ),
                    #  dfdp1 = p0 * dkdr * ( drdu * dudx + drdv * dvdx ) * dxdp1
                  1 : ( lambda: p0dkdr * ( drdu * c / params[3] + drdv * s / params[4] ) ),

                    #  dfdp2 = p0 * dkdr * ( drdu * dudy + drdv * dvdy ) * dydp2
                  2 : ( lambda: p0dkdr * ( -drdu * s / params[3] + drdv * c / params[4] ) ),

                    #  dfdp3 = p0 * dkdr * drdu * dudp3
                  3 : ( lambda: p0dkdr * drdu * u / params[3] ),
                  4 : ( lambda: p0dkdr * drdv * v / params[4] ),

                    #  dfdp5 = p0 * dkdr * ( drdu * ( dudc * dcdp5 + duds * dsdp5 ) +
                    #                        drdv * ( dvdc * dcdp5 + dvds * dsdp5 ) )
                  5 : ( lambda: -p0dkdr * ( drdu * ( -x * s - y * c ) / params[3] +
                                            drdv * ( -y * s + x * c ) / params[4] ) ) }

        if parlist is None :
            parlist = range( np )

        for k,kp in enumerate( parlist ) :
            partial[:,k] = parts[kp]()

        return partial





import numpy as numpy
import math
from . import Tools

from .Formatter import formatter as fmt
from .ConvergenceError import ConvergenceError

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2017        Do Kester

class AnnealingAmoeba( object ):
    """
    Simulated annealing simplex finding minimum.

    AnnealingAmoeba can be used in two modes: with simulated annealing on or off.
    The simulated annealing mode is invoked by setting the temperature to
    some value. By default it is off: temperature at zero.

    When the temperature is set at zero (default), AnnealingAmoeba acts as a
    simple Nelder-Mead downhill simplex method. With two advantages and one
    disadvantage. The pro's are that it is reasonably fast and that it does not
    need partial derivatives. The con is that it will fall into the first local
    minimum it encounters. No guarantee that this minimum has anything to do
    with the absolute minimum. This T=0 modus can only be used in mono-modal
    problems.

    In the other modus, when the temperature is set at some value the
    simplex sometimes takes a uphill step, depending on the temperature at
    that moment. Steps downhill are always taken. In that way it is possible
    to climb out of local minima to find better ones. Meanwhile the temperature
    is steadily lowered, concentrating the search on the by now hopefully found
    absolute minimum. Of course this takes much more iterations and still there
    is *no guarantee* that the best value is found. But a better chance.
    The initial temperature which suggests itself is of the order of the
    humps found in the minimizable lanscape.

    At each temperature level a number of moves is made. This number is set by
    the keyword steps=10, by default. After these steps the temperature is lowered
    with a factor set by cooling=0.95, by default.

    Iteration continues until the relative difference between the low and high
    points within the simplex is less than reltol
    ``|yhi - ylo| / ( |yhi| + |ylo| ) < reltol``
    and/or the absolute difference is less than abstol
    ``|yhi - ylo| < abstol``.

    AnnealingAmoeba can be used with limits set to one or more of the input values.

    The original version stems from Numerical Recipes with some additions of my own.

    Author       Do Kester

    Attributes
    ----------
    func : callable
        function to be minimized of form : `y = func( x )`
    lolimits : array_like
        lower limits on `x`. -inf is allowed to indicate no lower limit
    hilimits : array_like
        upper limits on `x`. +inf is allowed to indicate no upper limit
    fopt : float
        the best of the above values
    xopt : ndarray
        copy of the simplex point that has the best value (nx)

    rng : RandomState
        random number generator
    seed : int
        seed of rng

    reltol : float
        Relative tolerance. Program stops when ( |yhi-ylo| / (|yhi|+|ylo|) ) < reltol
    abstol : float
        Absolute tolerance. Program stops when |yhi-ylo| < abstol
    maxiter : int
        maximum number of iterations
    iter : int (read only)
        iteration counter
    ncalls : int (read only)
        numbers of calls to func

    temp : float
        annealing temperature (default: 0)
    cooling : float (non existent when temp=0)
        cooling factor (default: 0.95)
    steps : int (non existent when temp=0)
        number of steps per cooling cycle (default: 10)

    verbose : int
        0  : silent
        1 : print results to output
        2 : print some info every 100 iterations and plot results
        3 : print some info every iteration
    callback : callable
        function to be called every iteration of form :
        xopt = callback( xopt )

    simplex : ndarray
        the simplex has shape = (nx+1, nx); nx is the size of x
    values : ndarray
        the values of the function attained at the simplex points (nx+1).
    sum : ndarray
        sum over the corners of the simplex (nx)

    """
    #  *************************************************************************
    def __init__( self, func, xini, size=1, seed=4567, temp=0, limits=None,
                    maxiter=1000, reltol=0.0001, abstol=0.0001, cooling=0.95, steps=10,
                    verbose=0, callback=None ):
        """
        Create a new AnnealingAmoeba class to minimize the function

        Parameters
        ----------
        func : callable
            the function to be minimized
        xini : array_like
            initial values of the function
        size : float or array_like
            step size of the simplex
        seed : int
            for random number generator
        temp : float
            temperature of annealing (0 is no annealing)
        limits : None or list of 2 floats or list of 2 array_like
            None : no limits applied
            [lo,hi] : low and high limits for all values
            [la,ha] : low array and high array limits for the values
        maxiter : int
            max number of iterations
        reltol : float, None
            Relative tolerance. Program stops when ( |hi-lo| / (|hi|+|lo|) ) < reltol
        abstol : float, None
            Absolute tolerance. Program stops when |hi-lo| < abstol
            when abstol has a (float) value, reltol might be None.
        cooling : float
            cooling factor when annealing
        steps : int
            number of cycles in each cooling step.
        verbose : int
            0 : silent
            1 : print results to output
            2 : print some info every 100 iterations and plot results
            3 : print some info every iteration
        callback : callable
            is called each iteration as
            val = callback( val )
            where val is the minimizable array

        Raises
        ------
        ValueError
            1. When func is not callable
            2. When both tolerances are None
            3. When callback is not callable

        """
        super( AnnealingAmoeba, self ).__init__( )

        if callable( func ) :
            self.func = func
        else :
            raise ValueError( "func needs to be a callable function" )
        self.seed = seed
        self.rng = numpy.random.RandomState( self.seed )
        self.temp = temp
        nfit = Tools.length( xini )
        if Tools.length( size ) == 1 :
            size = [size] * nfit

        if limits is None :
            self.lolimits = None
            self.hilimits = None
        else :
            self.lolimits = limits[0] if Tools.length( limits[0] ) > 1 else [limits[0]]*nfit
            self.hilimits = limits[1] if Tools.length( limits[1] ) > 1 else [limits[1]]*nfit
        self.maxiter = maxiter
        if abstol is None and reltol is None :
            raise ValueError( "Both reltol and abstol are None." )
        self.abstol = abstol
        self.reltol = reltol
        if temp > 0 :
            self.cooling = cooling
            self.steps = steps
        self.verbose = verbose

        self.callback = callback
        if callback is not None and not callable( callback ) :
            raise ValueError( "callback needs to be a callable function (or None)" )

        # counters
        self.iter = 0
        self.ncalls = 0

        # initialize the simplex
        self.makeSimplex( xini, size )


    #  *************************************************************************
    def makeSimplex( self, xini, step ):
        """
        Make a simplex for the given set of parameters.

        Parameters  fitindex    list of parameters to fit

        """
        nfit = len( xini )

        i = 0
        for ir in xini :
            if self.hasLowLimits( i ) and ir <= self.lolimits[i] :
                lo = self.lolimits[i]
                if self.hasHighLimits( i ):
                    hi = self.hilimits[i]
                    xini[i] = 0.5 * ( hi + lo )
                    step[i] = 0.1 * ( hi - lo )
                else:
                    xini[i] = lo + 10
                    step[i] = 1
            if self.hasHighLimits( i ) and ir >= self.hilimits[i] :
                hi = self.hilimits[i]
                if self.hasLowLimits( i ):
                    lo = self.lolimits[i]
                    xini[i] = 0.5 * ( hi + lo )
                    step[i] = 0.1 * ( hi - lo )
                else:
                    xini[i] = hi - 10
                    step[i] = 1
            i += 1

        ndim = nfit + 1
        self.simplex = numpy.zeros( ( ndim, nfit ), dtype=float )
        self.values = numpy.zeros( ndim, dtype=float )
        self.xopt =  numpy.zeros( nfit, dtype=float )
        self.sum = numpy.zeros( nfit, dtype=float )

        for i in range( ndim ) :
            self.simplex[i,:] += xini
            if i < nfit :
                self.simplex[i,i] += step[i]
            trypar = self.stayInLimits( xini, self.simplex[i,:] )
            self.simplex[i,:] = trypar

        self.checkSimplex( self.simplex )
        self.setValues()

    def hasLowLimits( self, k ) :
        return self.lolimits is not None and self.hilimits[k] > -math.inf

    def hasHighLimits( self, k ) :
        return self.hilimits is not None and self.hilimits[k] < math.inf

    def stayInLimits( self, oldpar, trypar ) :
        if self.lolimits is None and self.hilimits is None :
            return trypar

        npar = self.simplex.shape[1]
        for k in range( npar ) :
            if self.lolimits[k] > trypar[k] :
                trypar[k] = ( oldpar[k] + self.lolimits[k] ) / 2
            elif trypar[k] > self.hilimits[k] :
                trypar[k] = ( oldpar[k] + self.hilimits[k] ) / 2
        return trypar

    def checkSimplex( self, simplex ):
        """
        Check for degeneracy: all points on same location.
        """
        dim = simplex.shape
        for i in range( dim[1] ) :
            same = True
            pre = simplex[0,i]
            same = all( sim == pre for sim in simplex[i,:] )

            if same:
                print( simplex[i,:] )
                raise ValueError( "Simplex is degenerate at index %d"%i )

    def setValues( self ):
        """
        Calculate the function values a simplex's corners

        """
        for i in range( self.simplex.shape[0] ):
            corner = self.simplex[i,:]
            cost = self.func( corner )
            if not numpy.isfinite( cost ) :
                corner += 0.01 * numpy.sum( self.simplex, 0 )
                cost = self.func( corner )

            self.values[i] = cost
            if i == 0 or self.values[i] < self.fopt:
                self.fopt = self.values[i]
                self.xopt = corner.copy()
        self.sum = numpy.sum( self.simplex, axis=0 )
        self.ncalls += self.simplex.shape[0]

# *************************************************************************
    def minimize( self ):
        """
        Converge the simplex.

        Returns
        -------
        ndarray : the optimal x values.

        Raises
        ------
        ConvergenceError when too many iterations are needed.
        """

        if self.verbose > 0 :
            print( "Iteration   step   ", ( "temp" if ( self.temp > 0 ) else "" ),
                   "  minimum    values" )
            numpy.set_printoptions( precision=3, suppress=True )

        maxiter = self.maxiter * ( 1 if self.temp == 0 else self.steps )
        while self.iter < maxiter :
#            print( self.iter, maxiter )
            trans = self.temperatureStep( )
            self.ncalls += trans

            if self.temp > 0 :
                self.temp *= self.cooling
            if trans == 0:
                return self.xopt

        raise ConvergenceError( "AnnealingAmoeba: Too many iterations: %d"%self.iter )

    def temperatureStep( self ):
        """
        Perform simplex moves in the right direction.

        Returns
        -------
        int : number of transforms.

        """
        ALPHA = 1.0
        BETA = 0.5
        GAMMA = 2.0
        DELTA = 10.

        (ndim,nfit) = self.simplex.shape
        maxiter = self.iter + ( self.maxiter if ( self.temp == 0 ) else self.steps * nfit )
        beta = BETA
        trans = 0

        self.sum = numpy.sum( self.simplex, axis=0 )
        move = "start"
        while self.iter < maxiter :
            ilo = 0
            ihi = 1
            ylo = self.values[0] + self.logRanTemp()
            ynhi = ylo
            yhi = self.values[1] + self.logRanTemp()

            if ylo > yhi:
                ihi = 0
                ilo = 1
                ynhi = yhi
                yhi = ylo
                ylo = ynhi
            for i in range( 2, ndim ) :
                yt = self.values[i] + self.logRanTemp()
                if yt <= ylo:
                    ilo = i
                    ylo = yt
                if yt > yhi:
                    ynhi = yhi
                    ihi = i
                    yhi = yt
                elif yt > ynhi:
                    ynhi = yt

#            for i in range( ndim ) :
#                print( self.simplex[i,:], self.values[i] )
#            print( ilo, ihi, ylo, ynhi, yhi )

            rtol = abs( yhi - ylo ) / max( abs( yhi ) + abs( ylo ), 1e-20 )
            stops = ( ( self.abstol is None or abs( yhi - ylo ) < self.abstol ) and
                      ( self.reltol is None or rtol < self.reltol ) )

#            print( rtol, self.reltol, abs( yhi - ylo ), self.abstol, stops, trans, self.fopt, ylo )

            if stops :
                # put best values and simplex in first position of the simplex
                if ilo > 0 :
                    swap = self.values[0]
                    self.values[0] = self.values[ilo]
                    self.values[ilo] = swap

                    swap = self.simplex[0,:]
                    self.simplex[0,:] = self.simplex[ilo,:]
                    self.simplex[ilo,:] = swap

                # check if best ever value is better than best in simplex;
                #     if so replace best and continue
                tol = self.abstol if self.abstol is not None else self.reltol
                if abs( self.fopt - ylo ) > tol :
                    self.simplex[ilo,:] = self.xopt[:]
                    self.values[ilo] = self.fopt

                    move = "replace"
                    if self.iter < maxiter:
                        continue

                if self.iter == 0 :             ## it does not start. Inflate the simplex
                    move = "expand"
                    ylo = self.inflateSimplex( ilo, DELTA )
                    trans += nfit
                    continue
                break

            trans += 2
            factor = -ALPHA
            while True:
                ytry = self.trialStep( ihi, yhi, factor )
                factor = self.randomRange( -ALPHA )
                if not math.isnan( ytry ) : break
            move = "reflect"					# reflect worst over the others

            if ytry < ylo:                      # when succesfull move further
                factor = GAMMA
                while True:
                    ytry = self.trialStep( ihi, yhi, factor )
                    factor = self.randomRange( GAMMA )
                    if not math.isnan( ytry ) : break
                move = "extrapolate"			# reflect twice as far

            elif ytry >= ynhi:                  # not better than next hi : contract
                ysave = yhi
                factor = beta
                while True:
                    ytry = self.trialStep( ihi, yhi, factor )
                    factor = self.randomRange( beta )
                    if not math.isnan( ytry ) : break
                move = "contract"				# contract in direction of others

                if ytry >= ysave :              # not better that saved one : shrink
                    factor = beta
                    while True:
                        ytry = self.inflateSimplex( ilo, factor )
                        factor = self.randomRange( beta )
                        if not math.isnan( ytry ) : break
                    trans += nfit
                    move = "shrink" 			# shrink simplex on best point
            else:
                trans -= 1

            if self.callback is not None :
                self.simplex[ihi,:] = self.callback( self.simplex[ihi,:] )

            self.iter += 1
            self.doVerbose( move, ytry, self.simplex[ihi,:], force=(self.verbose>=3) )

        self.doVerbose( move, self.values[0], self.simplex[0,:], force=True )
        return trans

    def doVerbose( self, name, chisq, par, force=False ):
        if self.verbose > 1 and ( self.iter % 100 == 0 or force ):
            mx = 5 if self.verbose < 4 else None if self.verbose == 4 else self.verbose
            print( "%6d    %-8.8s "%(self.iter,name),
                    ( "%6.1f"%(self.temp ) if ( self.temp > 0 ) else "" ),
                    " %8.1f  "%chisq, fmt( par, max=mx ) )

    def randomRange( self, factor ):
        return factor * ( 0.9 + 0.2 * self.rng.rand( ) )

    def inflateSimplex( self, ilo, factor ):
        """
        Inflate/deflate simplex around the (lowest) point (ilo).

        inflate if factor > 1
        deflate if factor < 1
        mirror  if factor < 0

        Parameters
        ----------
        ilo : int
            lowest point in the simplex
        factor : float
            inflation factor

        """
        emf = 1 - factor
        (ndim,nfit) = self.simplex.shape
        milo = [k for k in range( ilo )] + [k for k in range( ilo+1, ndim )]
        for i in milo :
            inpar = self.simplex[i,:]
            self.simplex[i,:] = factor * inpar + emf * self.simplex[ilo,:]

            self.simplex[i,:] = self.stayInLimits( inpar, self.simplex[i,:] )

        ylo = self.values[ilo]
        for i in milo :
            p = self.simplex[i,:]
            self.values[i] = self.func( p )
            if self.values[i] < ylo:
                ylo = self.values[i]
                if self.values[i] < self.fopt:
                    self.fopt = ylo
                    self.xopt = p

        self.sum = numpy.sum( self.simplex, axis=0 )

        return ylo

    def trialStep( self, ihi, yhi, factor ):
        """
        Do a trial step to improve the worst (highest) point.

        Parameters
        ----------
        ihi : int
            index of the high point
        yhi :  int
            value at the high point
        factor : int
            step size
        """
        (ndim,nfit) = self.simplex.shape

        fac1 = ( 1 - factor ) / nfit
        fac2 = fac1 - factor
        partry = self.sum * fac1 - self.simplex[ihi,:] * fac2
        inpar = self.simplex[ihi,:]
        partry = self.stayInLimits( inpar, partry )

        ytry = self.func( partry )

        if ytry < self.fopt:
            self.fopt = ytry
            self.xopt = partry.copy()

        yflu = ytry - self.logRanTemp( )

        if yflu < yhi:
            self.values[ihi] = ytry
            self.sum += partry - self.simplex[ihi,:]
            self.simplex[ihi,:] = partry

        return yflu

    def logRanTemp( self ):
        EPSILON = 1.0e-30
        return -self.temp * math.log( self.rng.rand( ) + EPSILON )


    def __str__( self ) :
        return "AnnealingAmoeba with temp = %f"%self.temp





import numpy as numpy
from astropy import units
import math
from . import Tools

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester SRON (Java code)
#  *    2016 - 2017 Do Kester

class ConvergenceError( Exception ):
    """
    Thrown when an iterative fitter stops while the minimum has not been found.

    Author:       Do Kester

    """
    pass





# run with : python3 -m unittest TestNLFitters

import unittest
import numpy as numpy
from numpy.testing import assert_array_almost_equal as assertAAE
from astropy import units
import math

import matplotlib.pyplot as plt

from BayesicFitting import *

from StdTests import stdFittertest

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *  2004 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2017 Do Kester

class TestNLFitters( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author:      Do Kester

    """
################################################################################
    def plotAmoebaFitter( self ):
        self.testAmoebaFitter( plot=True )
        plt.show()

    def testAmoebaFitter( self, plot=False ):
        options = {"temp": 10.0, "maxiter": 10000 }
        options = {"maxiter": 10000}
        stdFittertest( AmoebaFitter, 201, plot=plot, options=options )

    def testAmoebaFitter2( self, plot=False ):
        options = {"temp": 10.0, "maxiter": 10000, "verbose": 2 }
        stdFittertest( AmoebaFitter, 201, errdis='gauss', plot=plot, options=options )
        plt.show()

    def plotNelderMeadFitter( self ):
        self.testNelderMeadFitter( plot=True )
        plt.show()

    def testNelderMeadFitter( self, plot=False ):
        stdFittertest( NelderMeadFitter, 201, plot=plot, options={"maxiter": 3000} )

    def plotPowellFitter( self ):
        self.testPowellFitter( plot=True )
        plt.show()

    def testPowellFitter( self, plot=False ):
        stdFittertest( PowellFitter, 201, plot=plot )

    def plotConjugateGradientFitter( self ):
        self.testConjugateGradientFitter( plot=True )
        plt.show()

    def testConjugateGradientFitter( self, plot=False ):
        stdFittertest( ConjugateGradientFitter, 201, plot=plot )

    def plotBfgsFitter( self ):
        self.testBfgsFitter( plot=True )
        plt.show()

    def testBfgsFitter( self, plot=False ):
        stdFittertest( BfgsFitter, 201, plot=plot )

    def plotNewtonCgFitter( self ):
        self.testNewtonCgFitter( plot=True )
        plt.show()

    def testNewtonCgFitter( self, plot=False ):
        stdFittertest( NewtonCgFitter, 201, plot=plot )

    def plotLbfgsbFitter( self ):
        self.testLbfgsbFitter( plot=True )
        plt.show()

    def testLbfgsbFitter( self, plot=False ):
        stdFittertest( LbfgsbFitter, 201, plot=plot )

    def plotTncFitter( self ):
        self.testTncFitter( plot=True )
        plt.show()

    def testTncFitter( self, plot=False ):
        stdFittertest( TncFitter, 201, plot=plot )

    def plotCobylaFitter( self ):
        self.testCobylaFitter( plot=True )
        plt.show()

    def testCobylaFitter( self, plot=False ):
        stdFittertest( CobylaFitter, 201, plot=plot )

    def plotSlsqpFitter( self ):
        self.testSlsqpFitter( plot=True )
        plt.show()

    def testSlsqpFitter( self, plot=False ):
        stdFittertest( SlsqpFitter, 201, plot=plot )

    def plotDoglegFitter( self ):
        self.testDoglegFitter( plot=True )
        plt.show()

    def testDoglegFitter( self, plot=False ):
        stdFittertest( DoglegFitter, 201, plot=plot )

    def plotTrustNcgFitter( self ):
        self.testTrustNcgFitter( plot=True )
        plt.show()

    def testTrustNcgFitter( self, plot=False ):
        stdFittertest( TrustNcgFitter, 201, plot=plot )

    def plotall( self ):
        self.testAmoebaFitter( plot=True )
        self.testNelderMeadFitter( plot=True )
        self.testPowellFitter( plot=True )
        self.testConjugateGradientFitter( plot=True )
        self.testBfgsFitter( plot=True )
        self.testNewtonCgFitter( plot=True )
        self.testLbfgsbFitter( plot=True )
        self.testTncFitter( plot=True )
        self.testCobylaFitter( plot=True )
        self.testSlsqpFitter( plot=True )
        self.testDoglegFitter( plot=True )
        self.testTrustNcgFitter( plot=True )
        plt.show()


    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestNLFitters.__class__ )





import numpy as numpy
from . import Tools
from .NonLinearModel import NonLinearModel

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2007 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2017 Do Kester

class ExpModel( NonLinearModel ):
    """
    Exponential Model.
    .. math::
        f( x:p ) = p_0 * exp( p_1 * x )

    where p_0 = amplitude, p_1 = slope and
    As always x = input.

    The parameters are initialized at {1.0, -1.0}. It is a non-linear model.

    Beware of a positive 2nd parameter; when positive the model is going off
    to Infinity quite quickly.


    Examples
    --------
    >>> em = ExpModel( )
    >>> print( em.getNumberOfParameters( ) )
    2

    Category:    mathematics/Fitting

    """
    def __init__( self, copy=None, **kwargs ):
        """
        Exponential model.
        <br>
        Number of parameters is 2.

        Parameters
        ----------
        copy : ExpModel
            to be copied

        """
        param = [1.0,-1.0]
        names = ["amplitude","slope"]
        super( ExpModel, self ).__init__( 2, copy=copy, params=param,
                names=names, **kwargs )

    def copy( self ):
        """ Copy method.  """
        return ExpModel( copy=self )

    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the result
        params : array_like
            values for the parameters

        """
        return numpy.multiply( numpy.exp( numpy.multiply( params[1], xdata ) ), params[0] )

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the result
        params : array_like
            values for the parameters
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        np = self.npbase if parlist is None else len( parlist )
        partial = numpy.ndarray( ( Tools.length( xdata ), np ) )

        e = numpy.exp( params[1] * xdata )

        parts = { 0 : ( lambda: e ),
                  1 : ( lambda: params[0] * e * xdata ) }

        if parlist is None :
            parlist = range( self.npmax )

        for k,kp in enumerate( parlist ) :
            partial[:,k] = parts[kp]()

        return partial

    def baseDerivative( self, xdata, params ):
        """
        Returns the derivative df/dx at the input value.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the result
        params : array_like
            values for the parameters

        """
        return numpy.multiply( params[1], self.baseResult( xdata, params ) )

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        return str( "Exp: f( x:p ) = p_0 * exp( p_1 * x )" )

    def baseParameterUnit( self, k ):
        """
        Return the unit of the indicated parameter.
        Parameters: k    parameter number.

        """
        if k == 0:
            return self.yUnit
        return self.xUnit ** ( -1 )





from __future__ import print_function

import numpy as numpy
from astropy import units
import math
from . import Tools
from .Formatter import formatter as fmt
from . import Plotter
import sys
import warnings
import matplotlib.pyplot as plt

from .Model import Model
from .Sample import Sample
from .SampleList import SampleList
from .ErrorDistribution import ErrorDistribution
from .ScaledErrorDistribution import ScaledErrorDistribution
from .GaussErrorDistribution import GaussErrorDistribution
from .LaplaceErrorDistribution import LaplaceErrorDistribution
from .PoissonErrorDistribution import PoissonErrorDistribution
from .CauchyErrorDistribution import CauchyErrorDistribution
from .GenGaussErrorDistribution import GenGaussErrorDistribution
from .Explorer import Explorer
from .Engine import Engine
from .StartEngine import StartEngine
from .GibbsEngine import GibbsEngine
from .GalileanEngine import GalileanEngine
from .StepEngine import StepEngine


## for Dynamic Models import the classes
from .BirthEngine import BirthEngine
from .DeathEngine import DeathEngine

#######  For later #################
## for Order Problems import the classes
#from OrderProblem import OrderProblem
#from DistanceCostFunction import DistanceCostFunction
#from StartOrderEngine import StartOrderEngine
#from OrderEngine import OrderEngine
#from ReverseEngine import ReverseEngine
#from ShuffleEngine import ShuffleEngine
#from SwitchEngine import SwitchEngine


__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2017        Do Kester

class NestedSampler( object ):
    """
    A novel technique to do Bayesian calculation.

    NestedSampler produces `Sample`s from the posterior probability
    distribution of the parameters, given a `Model` and a
    dataset ( x-values, y-values and optionally weights ).
    The Samples are collected in a `SampleList`.

    NestedSampler is constructed according to the ideas of John Skilling
    and David MacKay ( ref TBD ).

    Internally it contains an ensemble ( by default: 100 ) of trial samples,
    called walkers. Initially they are randomly distributed over the space
    available to the parameters. This randomness is distributed according
    to the prior distribution of the parameters. In most simple cases it
    will be uniform.

    In an iterative process, the sample with the lowest likelihood is
    selected and replaced by a copy of one of the others.
    The parameters of the copy are randomly walked around under the
    condition that its likelihood stays larger than the selected lowest likelihood.
    A new independent trial sample is constructed.
    The original lowest sample is placed, appropriately weighted, into the
    SampleList for output.

    This way the likelihood is climbed until the maximum is found.
    Along the way the integral of the likelihood function is
    calculated, which is equal to the evidence for the resulting parameters,
    given the dataset.

    There are 5 different likelihood functions available: Gaussian (Normal),
    Laplace (Exponential), Poisson (for counting processes),
    Cauchy (aka Lorentz) and a generalized Gaussian.
    By default the Gaussian distribution is used.

    Except for Poisson, all others are `ScaledErrorDistribution`s. The scale is
    a `HyperParameter` which can either be fixed, by setting the attribute
    errdis.scale or optimized, given the model parameters and the data.
    By default it is optimized for Gaussian and Laplace distributions.
    The prior for the noise scale is a Jeffreys distribution.

    The generalized Gaussian also has a power. It is also a `HyperParameter`
    which can be fixed or optimized. Its default is 2 (==Gaussian).

    For the random walk of the parameters 4 so-called engines are written.
    By default only he first is switched on.

    GalileanEngine. It walks all (super)parameters in a fixed direction
        for about 5 steps. When a step ends outside the high likelihood
        region the direction is mirrored on the lowLikelihood edge and
        continued.

    GibbsEngine. It moves each of the parameters by a random step.
        It is a randomwalk.

    StepEngine. It moves all parameters in a random direction.
        It is a randomwalk.

    CrossEngine. The parameters of 2 walkers are mixed in a random way.

    For dynamic models 2 extra engines are defined:

    BirthEngine. It tries to increase the number of parameters.
    DeathEngine. It tries to decrease the number of parameters.

    Attributes
    ----------
    xdata : array_like
        array of independent input values
    model : Model
        the model function to be fitted
    ydata : array_like
        array of dependent (to be fitted) data
    weights : array_like (None)
        weights pertaining to ydata
    distribution : ErrorDistribution
        to calculate the loglikelihood
    ensemble : int (100)
        number of walkers
    discard : int (1)
        number of walkers to be replaced each generation
    rng : RandomState
        random number generator
    seed : int (80409)
        seed of rng
    rate : float (1.0)
        speed of exploration
    maxsize : None or int
        maximum size of the resulting sample list (None : no limit)
    end : float (2.0)
        stopping criterion
    verbose : int
        level of blabbering

    walkers : SampleList
        ensemble of Samples that explore the likelihood space
    samples : SampleList
        Samples resulting from the exploration
    engines : list of Engine
        Engine that move the walkers around within the given constraint: logL > lowLogL
    initialEngine : Engine
        Engine that distributes the walkers over the available space
    restart : StopStart (TBW)
        write intermediate results to (optionally) start from.


    Author       Do Kester.


    """
    TWOP32 = 2 ** 32


    #  *********CONSTRUCTORS***************************************************
    def __init__( self, xdata, model, ydata, weights=None, distribution=None,
                keep=None, ensemble=100, discard=1, seed=80409, rate=1.0,
                limits=None, engines=None, maxsize=None, verbose=1 ) :
        """
        Create a new class, providing inputs and model.

        Parameters
        ----------
        xdata : array_like
            array of independent input values
        model : Model
            the model function to be fitted
            the model needs priors for the parameters and (maybe) limits
        ydata : array_like
            array of dependent (to be fitted) data
        weights : array_like (None)
            weights pertaining to ydata
        keep : None or dict of {int:float}
            None : none of the model parameters are kept fixed.
            Dictionary of indices (int) to be kept at a fixed value (float).
            Hyperparameters follow model parameters.
            The values will override those at initialization.
            They are used in this instantiation, unless overwritten at the call to sample()
        distribution : None or String or ErrorDistribution
            None   : GaussErrorDistribution is chosen with (fixed) scale equal to 1.0

            "gauss"    : `GaussErrorDistribution`    with 1 hyperpar scale
            "laplace"  : `LaplaceErrorDistribution`  with 1 hyperpar scale
            "poisson"  : `PoissonErrorDistribution`  no hyperpar
            "cauchy"   : `CauchyErrorDstribution`    with 1 hyperpar scale
            "gengauss" : `GenGaussErrorDistribution` with 2 hyperpar (scale, power)
            Only for OrderProblems: (for later)
            "distance" : `DistanceCostFunction`      no hyperpar

            Error distribution class which implements logLikelihood

            When the hyperpar(s) are not to be kept fixed, they need `Prior` and maybe limits.
        ensemble : int (100)
            number of walkers
        discard : int (1)
            number of walkers to be replaced each generation
        seed : int (80409)
            seed of rng
        rate : float (1.0)
            speed of exploration
        engines : None or (list of) string or (list of) Engine
            to randomly move the walkers around, within the likelihood bound.

            "galilean"  : GalileanEngine
            "gibbs" 	: GibbsEngine 	move one parameter at a time
            "step"  	: StepEngine    move all parameters in arbitrary direction

            For OrderProblems (for later)
            "order"   : insert a snippet of parameters at another location
            "reverse" : reverse the order of a snippet of parameters

            For Dynamic models only
            "birth" : BirthEngine   increase the parameter list of a walker by one
            "death" : DeathEngine   decrease the parameter list of a walker by one

            None    : take default ["galilean"]. When the model is Dynamic
                      the list is supplemented with ["birth", "death"]

            engine  : a class inheriting from Engine. At least implementing
                      execute( walker, lowLhood, fitIndex )
        maxsize : None or int
            maximum size of the resulting sample list (None : no limit)
        vebose : int
            0 : silent
            1 : basic information
            2 : more about every 100th iteration
            3 : more about every iteration

        """
        self.xdata = xdata
        self.model = model
        if not ( self.model.priors and all( p.isBound() for p in self.model.priors ) ) :
            warnings.warn( "Model needs priors and/or limits" )
        self.ydata = ydata
        self.weights = weights
        self.keep = keep

        self.ensemble = ensemble
        self.discard = discard
        self.rng = numpy.random.RandomState( seed )
        self.seed = seed
        self.maxsize = maxsize
        self.verbose = verbose
        self.rate = rate
        self.restart = None                     ## TBD

        self.minimumIterations = 100
        self.end = 2.0
        self.maxtrials = 5

        self.iteration = 0

        if distribution is not None :
            self.setErrorDistribution( distribution )
#        elif isinstance( model, OrderProblem ) :
#            self.setErrorDistribution( "distance" )
        else :
            self.setErrorDistribution( "gauss", scale=1.0 )

#        print( distribution.hyperpar[0].getLimits() )
#        print( self.distribution.hyperpar[0].getLimits() )

        self.setEngines( engines )

        ## Initialize the sample list
        self.samples = SampleList( model, 0, self.distribution, ndata=len( ydata ) )

    def makeFitlist( self, keep=None ) :
        """
        Make list of indices of (hyper)parameters that need fitting.

        Parameters
        ----------
        keep : None or dict of {int : float}
            dictionary of indices that need to be kept at the float value.
        """

        np = len( self.model.parameters )
        fitlist = [k for k in range( np )]
        nh = -1
        for sp in self.distribution.hyperpar :
            if not sp.isFixed and sp.isBound() :
                fitlist += [nh]
            nh -= 1

        if keep is not None :
            fitlist = numpy.setxor1d( fitlist, list( keep.keys() ) )
            self.model.parameters[list(keep.keys())] = list( keep.values() )

        return fitlist


    #  *******SAMPLE************************************************************
    def sample( self, keep=None, plot=False ):
        """
        Sample the posterior and return the weighted average result of the
        Model.

        The more sensible result of this method is a SampleList which contains
        samples taken from the posterior distribution.

        Parameters
        ----------
        keep : None or dict of {int:float}
            Dictionary of indices (int) to be kept at a fixed value (float)
            Hyperparameters follow model parameters
            The values will override those at initialization.
            They are only used in this call of fit.
        plot : bool
            Show a plot of the results

        """
        if keep is None :
            keep = self.keep
        fitlist = self.makeFitlist( keep=keep )

        self.initWalkers( fitlist=fitlist )

        for eng in self.engines :
            eng.walkers = self.walkers

        self.distribution.ncalls = 0                      #  reset number of calls

        self.plotData( )

        if self.verbose >= 1 :
            print( "Fit", ( "all" if keep is None else fitlist ), "parameters of" )
            print( " ", self.model._toString( "  " ) )
            print( "Using a", self.distribution, "with", end="" )
            np = -1
            cstr = " with "
            for name,hyp in zip( self.distribution.PARNAMES, self.distribution.hyperpar ) :
                print( cstr, end="" )
                if np in fitlist :
                    print( "unknown %s" % name, end="" )
                else :
                    print( "%s = %7.2f " % (name, hyp.hypar), end="" )
                np -= 1
                cstr = " and "
            print( "\nMoving the walkers with ", end="" )
            for eng in self.engines :
                print( " ", eng, end="" )
            print( "" )


        if self.verbose >= 2:
            print( "Iteration   logZ        H     LowL     npar    parameters" )

        explorer = Explorer( self )

        self.logZ = -sys.float_info.max
        self.info = 0

        logWidth = math.log( 1.0 - math.exp( (-1.0 * self.discard ) / self.ensemble) )

#        for w in self.walkers :
#            print( w.id, w.allpars, w.logL )

        if self.optionalRestart() :
            logWidth -= self.iteration * ( 1.0 * self.discard ) / self.ensemble

        while self.iteration < self.getMaxIter( ):

            #  find worst walker(s) in ensemble
            worst = self.findWorst()
            worstLogW = logWidth + self.walkers[worst[-1]].logL

            # Keep posterior samples
            self.storeSamples( worst, worstLogW - math.log( self.discard ) )

            # Update Evidence Z and Information H
            logZnew = numpy.logaddexp( self.logZ, worstLogW )

            self.info = ( math.exp( worstLogW - logZnew ) * self.lowLhood +
                    math.exp( self.logZ - logZnew ) * ( self.info + self.logZ ) - logZnew )
            self.logZ = logZnew

            if self.verbose >= 3 or ( self.verbose >= 2 and self.iteration % 100 == 0 ):
                kw = worst[0]
                pl = self.walkers[kw].allpars[self.walkers[kw].fitIndex]
                np = len( pl )
                print( "%8d %8.1f %8.1f %8.1f %6d "%( self.iteration, self.logZ, self.info,
                        self.lowLhood, np ), fmt( pl ) )

                self.plotResult( worst[0], self.iteration )

            self.samples.weed( self.maxsize )                # remove overflow in samplelist

            self.copyWalker( worst )

            # Explore the copied walker(s)
#            print( "NS     ", self.walkers[worst[0]].allpars, fitlist )
            explorer.explore( worst, self.lowLhood, fitlist )

            # Shrink the interval
            logWidth -= ( 1.0 * self.discard ) / self.ensemble
            self.iteration += 1

            self.optionalSave( )


        # End of Sampling
        self.addEnsembleToSamples( logWidth )

        # Calculate weighted average and stdevs for the parameters;
        self.samples.LogZ = self.logZ
        self.samples.info = self.info
        self.samples.normalize( )

        # put the info into the model
        if not self.model.isDynamic() :
            self.model.parameters = self.samples.parameters
            self.model.stdevs = self.samples.stdevs

        if self.verbose >= 1 :
            self.report()

#        for eng in self.engines :
#            print( eng )
#            print( " s  ", eng.succ )
#            print( " f  ", eng.fail )

        yfit = self.samples.average( self.xdata )
        if plot :
            Plotter.plotFit( self.xdata, self.ydata, yfit=yfit,
                             residuals=True )

        return yfit

    def getMaxIter( self ) :
        return max( self.minimumIterations, self.end * self.ensemble * self.info / self.discard )


#  ===================================================================================
    def optionalRestart( self ):

        if self.restart is not None and self.restart.wantRestore( ):
            self.walkers, self.samples = self.restart.restore( self.walkers, self.samples )
            self.logZ = self.walkers.logZ
            self.info = self.walkers.info
            self.iteration = self.walkers.iteration
            return True
        return False

    def optionalSave( self ):
        if self.restart is not None and self.restart.wantSave( ) and self.iteration % 100 == 0:
            self.walkers.logZ = self.logZ
            self.walkers.info = self.info
            self.walkers.iteration = self.iteration
            self.restart.save( self.walkers, self.samples )

    def storeSamples( self, worst, worstLogW ):
        for kw in worst :
            self.walkers[kw].logW = worstLogW
            self.samples.add( self.walkers, kw )

    def findWorst( self ):
        """
        Find discard bad points in ensemble. In order worse to better.
        lowLhood is the "best" in the bad points.

        """
        worst = []
        for k in range( self.discard ) :
            self.lowLhood = sys.float_info.max
            for i in range( self.ensemble ):
                if self.walkers[i].logL < self.lowLhood :
                    if i in worst :
                        continue
                    bad = i
                    self.lowLhood = self.walkers[i].logL
            worst += [bad]
        return worst

    def copyWalker( self, worst ):
        """
        Kill worst walker( s ) in favour of one of the others

        """
        sworst = sorted( worst )
#        print( worst, sworst )
        for k in range( self.discard ):
            kcp = self.rng.randint( 0, self.ensemble - self.discard )
            for kk in range( self.discard ):
                if kcp >= sworst[kk] :
                    kcp += 1
            self.walkers.copy( kcp, worst[k] )

    def addEnsembleToSamples( self, logWidth ):
        """
        Add the ensemble walkers to the samples
        """
        done = [False] * self.ensemble

        for nest in range( self.ensemble ):
            # skip walkers already done
            worst = 0
            while done[worst] : worst += 1

            #  find worst walker in ensemble not yet handled
            self.lowLhood = self.walkers[worst].logL
            for i in range( worst+1, self.ensemble ) :
                logl = self.walkers[i].logL
                if logl < self.lowLhood and not done[i]:
                    worst = i
                    self.lowLhood = logl

            worstLogW = logWidth + self.lowLhood
            self.walkers[worst].logW = worstLogW

            # Update Evidence Z and Information H
            logZnew = numpy.logaddexp( self.logZ, worstLogW )
            self.info = ( math.exp( worstLogW - logZnew ) * self.lowLhood +
                math.exp( self.logZ - logZnew ) * ( self.info + self.logZ ) - logZnew )
            self.logZ = logZnew

            # Keep posterior sample
            self.samples.add( self.walkers, worst )

            done[worst] = True


    #  *********INTERNALS***************************************************
    def __setattr__( self, name, value ) :

        ### TBD hypar fitting
        if name == "scale" and isinstance( self.distribution, ScaledErrorDistribution ) :
            self.distribution.scale = value
        elif name == "power" and isinstance( self.distribution, GenGaussErrorDistribution ) :
            self.distribution.power = value
        else :
            object.__setattr__( self, name, value )

    def __getattr__( self, name ) :
        if name == "evidence" :
            return self.logZ / math.log( 10.0 )
        if name == "logZprecision" :
            return math.sqrt( self.info * self.discard / self.ensemble )
        if name == "precision" :
            return self.logZprecision / math.log( 10.0 )
        if name == "information" :
            return self.info

        if name == "parameters" :
            return self.samples.getParameters()
        elif name == "stdevs" or name == "standardDeviations" :
            self.samples.getParameters()
            return self.samples.stdevs
        elif name == "hypars" :
            return self.samples.hypars
        elif name == "stdevHypars" :
            return self.samples.stdevHypars
        elif name == "scale" :
            return self.samples.hypars[0]
        elif name == "stdevScale" :
            return self.samples.stdevHypars[0]

    #  *********DISTRIBUTIONS***************************************************
    def setErrorDistribution( self, name, scale=1.0, power=2.0 ):
        """
        Set the error distribution for calculating the likelihood.

        Parameters
        ----------
        name : string
            name of distribution: "gauss", "laplace", "poisson", "cauchy", "gengauss"
        scale : float
            fixed scale of distribution
        power : float
            fixed power of distribution

        """
        if isinstance( name, ErrorDistribution ) :
            self.distribution = name
            self.distribution.xdata = self.xdata
            self.distribution.data  = self.ydata
            self.distribution.weights = self.weights
            return

        if not isinstance( name, str ) :
            raise ValueError( "Cannot interpret ", name, " as string or ErrorDistribution" )

        name = str.lower( name )
#        print( name )
        if name == "gauss" :
            self.distribution = GaussErrorDistribution( self.xdata, self.ydata,
                    weights=self.weights, scale=scale )
        elif name == "laplace" :
            self.distribution = LaplaceErrorDistribution( self.xdata, self.ydata,
                    weights=self.weights, scale=scale )
        elif name == "poisson" :
            self.distribution = PoissonErrorDistribution( self.xdata, self.ydata,
                    weights=self.weights )
        elif name == "cauchy" :
            self.distribution = CauchyErrorDistribution( self.xdata, self.ydata,
                    weights=self.weights, scale=scale )
        elif name == "gengauss" :
            self.distribution = GenGaussErrorDistribution( self.xdata, self.ydata,
                    weights=self.weights, scale=scale, power=power )
        elif name == "distance" :
            self.distribution = DistanceCostFunction( self.xdata, self.ydata,
                    weights=self.weights )
        else :
            raise ValueError( "Unknown error distribution %s" % name )

    def setEngines( self, engines ) :
        """
        initialize the engines.

        Parameters
        ----------
        engines : list of string
            list of engine names
        """
        if engines is not None :
            pass
#        elif isinstance( self.model, OrderProblem ) :
#            engines = ["order", "reverse", "shuffle", "switch"]
        else :
            engines = ["galilean"]

        if self.model.isDynamic() :
            engines += ["birth", "death"]


        self.engines = []
        if isinstance( engines, str ) :
            engines = [engines]
        for name in engines :
            if isinstance( name, Engine ) :
                engine = name
                engine.members = self.walkers
                engine.errdis = self.distribution
                self.engines += [engine]
                continue

            if not isinstance( name, str ) :
                raise ValueError( "Cannot interpret ", name, " as string or as Engine" )


            if name == "galilean" :
                seed = self.rng.randint( self.TWOP32 )
#                print( "Galil    ", seed )
                engine = GalileanEngine( self.walkers, self.distribution, seed=seed )
            elif name == "gibbs" :
                seed = self.rng.randint( self.TWOP32 )
                engine = GibbsEngine( self.walkers, self.distribution, seed=seed )
            elif name == "step" :
                seed = self.rng.randint( self.TWOP32 )
                engine = StepEngine( self.walkers, self.distribution, seed=seed )
####        Order Problems (for later)
#            elif name == "order" :
#                seed = self.rng.randint( self.TWOP32 )
#                engine = OrderEngine( self.walkers, self.distribution, seed=seed )
#            elif name == "reverse" :
#                seed = self.rng.randint( self.TWOP32 )
#                engine = ReverseEngine( self.walkers, self.distribution, seed=seed )
#            elif name == "shuffle" :
#                seed = self.rng.randint( self.TWOP32 )
#                engine = ShuffleEngine( self.walkers, self.distribution, seed=seed )
#            elif name == "switch" :
#                seed = self.rng.randint( self.TWOP32 )
#                engine = SwitchEngine( self.walkers, self.distribution, seed=seed )

####        For Dynamic Models
            elif name == "birth" :
                seed = self.rng.randint( self.TWOP32 )
#                print( "Birth    ", seed )
                engine = BirthEngine( self.walkers, self.distribution, seed=seed )
            elif name == "death" :
                seed = self.rng.randint( self.TWOP32 )
#                print( "Death    ", seed )
                engine = DeathEngine( self.walkers, self.distribution, seed=seed )
            else :
                raise ValueError( "Unknown Engine name : %10s" % name )

            self.engines += [engine]


    #  *********INITIALIZATION***************************************************
    def initWalkers( self, fitlist=None ):
        """
        Initialize the walkers at random values of parameters and scale
        """
        self.walkers = SampleList( self.model, self.ensemble, self.distribution,
                fitindex=fitlist )

        if self.initialEngine is not None:
            # decorate with proper information
            self.initialEngine.members = self.walkers
            self.initialEngine.errdis = self.distribution
#        elif isinstance( self.model, OrderProblem ) :
#            seed = self.rng.randint( self.TWOP32 )
#            self.initialEngine = StartOrderEngine( self.walkers, self.distribution,
#                        seed=seed )
        else :
            seed = self.rng.randint( self.TWOP32 )
#            print( "Start    ", seed )
            self.initialEngine = StartEngine( self.walkers, self.distribution,
                        seed=seed )

        for walker in self.walkers :
            self.initialEngine.execute( walker, 0, fitIndex=fitlist )


    def plotData( self ):
        if self.plotter is None:
            return
        inp = self.xdata.getNumericData( )
        self.plotter.plotData( inp, self.data, "Iteration plot" )

    def plotResult( self, walkerId, iter ):
        if self.plotter is None:
            return
        inp = self.xdata.getNumericData( )
        param = self.walkers.getParameters( walkerId )
        model = self.walkers.getModel( walkerId )
        mock = model.result( self.xdata, param )
        self.plotter.plotResult( inp, mock, iter )

    def report( self ):

#        print( "Rate        %f" % self.rate )
        print( "Engines              success     reject     failed      calls" )

        for engine in self.engines :
            print( "%-16.16s " % engine, end="" )
            engine.printReport()
        print( "Calls to LogL     %10d" % self.distribution.ncalls, end="" )
        if self.distribution.nparts > 0 :
            print( "   to dLogL %10d" % self.distribution.nparts )
        else :
            print( "" )

        print( "Samples  %10d" % len( self.samples ) )
        print( "Evidence    %10.3f +- %10.3f" % (self.evidence, self.precision ) )






from __future__ import print_function

import numpy as numpy
from numpy import ndarray
from numpy import float64
from numpy import int64
import math

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2017        Do Kester


gllen = 80
gindent = 0
gmax = 5
fmt = { "float64" : " %8.3f", "int64" : " %8d" }


def formatter_init( format={}, indent=None, linelength=None, max=-1 ):
    """
    Initialize the formatter with new default values.

    Parameters
    ----------
    format : dict {namestring : formatstring }
        name : "float64" or "int64"
        fmt  : " %fmt"
    indent : None or int
        number of spaces to indent *after* the first line
        Default is 0
    linelength : None or int
        length of the lines produced
        Default is 80
    max : None or int
        maximum number of items displayed followed by ... if there are more
        None displays all
        Default is 5
    """

    global gllen, gindent, gmax

    for k in format.keys() :
        fmt[k] = format[k]

    if indent is not None :
        gindent = indent
    if linelength is not None :
        gllen = linelength
    if max is None or max >= 0 :
        gmax = max
#    print( "init  ", gindent, gllen, gmax, fmt )


def formatter( array, format=None, indent=None, linelength=None, max=-1 ) :
    """
    Format a number or an array nicely into a string

    Parameters override defaults given earlier with init().

    Parameters
    ----------
    array : array_like or number
        number or list of numbers or n-dim array of numbers
    format : None or string
        format applying to one item of array
        Default is "8.3f" for float and "8d" for int
    indent : None or int
        number of spaces to indent *after* the first line
        Default is 0
    linelength : None or int
        length of the lines produced
        Default is 80
    max : None or int
        maximum number of items displayed followed by ... if there are more
        None displays all
        Default is 5

    Returns
    -------
    string : containing the formatted array

    """
    global count, nwl, sp, llen, fmtlen, mx, fmt

    if indent is None :
        indent = gindent
    if linelength is None :
        linelength = gllen
    if max is not None and max <= 0 :
        max = gmax
    count = indent
    dent = indent
    llen = linelength
    mx = max
    nwl = False
    sp = 0
#    print( count, nwl, sp, llen, indent, mx, fmt )

    if not isinstance( array, ndarray ) :
        array = numpy.asarray( array )

    if format is None :
        format = fmt[str(array.dtype)]

    fmtlen = len( format % 1 )

    result = ""
    result = recursive_format( result, array, format=format, indent=indent )
    return result

def recursive_format( result, array, format=None, indent=0 ) :
    global count, nwl, sp, llen, fmtlen, mx, fmt

    if array.size == 1 :
        if count + fmtlen > llen :
            result += ( "\n%s" % spaces( sp + indent ) )
            count = indent
        if format is None :
            format = fmt[array.__class__]
        result += ( format % array )
        count += fmtlen
        return result

    shp = array.shape
    if sp > 0 and nwl :
        result += ( "\n%s" % spaces( sp + indent ) )
    result += ( "[" )
    nwl = False
    sp += 1
    shp0 = shp[0] if mx is None else min( shp[0], mx )
    for k in range( shp0 ) :
        result = recursive_format( result, array[k], format=format, indent=indent )
        nwl = True
    if mx is not None and shp[0] > mx :
        result += " ..."
    result += ( "]"  )
    count = 0
    sp -=1
    return result

def spaces( int ) :
    return ( " " * int )




# run with : python3 -m unittest TestNestedSampler
# or :       python3 -m unittest TestNestedSampler.Test.test1

import unittest
import numpy as numpy
from astropy import units
import math
from numpy.testing import assert_array_almost_equal as assertAAE
from FitPlot import plotFit

from BayesicFitting import *
from BayesicFitting import formatter as fmt


__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class TestNestedSampler( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author       Do Kester

    """

    def makeData( self, n=3 ) :

        N = 201
        a0 = 8.0
        x0 = 0.7
        s0 = 0.4
        b0 = 1.0
        b1 = 0.2
        p0 = 1.3
        cs = 0.3
        cc = 0.2
        r0 = 0.1


        x = numpy.arange( N, dtype=float ) / 25 - 2
        w = numpy.full( N, 2.0, dtype=float )

        g = ( x - x0 ) / s0
        g *= -0.5 * g
        g = a0 * numpy.exp( g )

        b = b1 * x + b0

        s = 2 * math.pi * p0 * x
        s = cs * numpy.sin( s ) + cc * numpy.cos( s )

        numpy.random.seed( 13456 )

        r = r0 * numpy.random.randn( N )
        if n == 1 :
            y0 = g
            pp = numpy.asarray( [a0,x0,s0], dtype=float )
        elif n == 2 :
            y0 = g + b
            pp = numpy.asarray( [a0,x0,s0,b0,b1], dtype=float )
        else :
            y0 = g + b + s
            pp = numpy.asarray( [a0,x0,s0,b0,b1,p0,cc,cs], dtype=float )

        y = y0 + r

        return pp, y0, x, y, w


    def plot1( self ):
        self.test1( plot=True )

    def plot2( self ):
        self.test2( plot=True )

    def plot3( self ):
        self.test3( plot=True )


    def test1( self, plot=False ):
        print( "=========== Nested Sampler test 1 ======================" )

        pp, y0, x, y, w = self.makeData( n=1 )

        gm = GaussModel( )

        print( gm.shortName( ) )

        lolim = numpy.asarray( [-10,-10,  0], dtype=float )
        hilim = numpy.asarray( [ 10, 10, 10], dtype=float )

        gm.setLimits( lolim, hilim )

        ns = NestedSampler( x, gm, y, w )
        ns.verbose = 2

        self.dofit( ns, pp )

        if plot :
            plotFit( x, y, gm, ftr=ns.samples )

    def test2( self, plot=False ):
        print( "=========== Nested Sampler test 2 ======================" )

        pp, y0, x, y, w = self.makeData( 2 )

        gm = GaussModel( )
        gm.addModel( PolynomialModel(1) )

        print( gm.shortName( ) )
        print( gm._next.shortName( ) )

        lolim = numpy.asarray( [-10,-10,  0,-10,-10], dtype=float )
        hilim = numpy.asarray( [ 10, 10, 10, 10, 10], dtype=float )

        gm.setLimits( lolim, hilim )

        ns = NestedSampler( x, gm, y )
        ns.verbose = 2

        print( "truth  ", pp )
        self.dofit( ns, pp )

        if plot :
            plotFit( x, y, gm, ftr=ns.samples )

    def test3( self, plot=False ):
        print( "=========== Nested Sampler test 3 ======================" )

        pp, y0, x, y, w = self.makeData( 3 )

        gm = GaussModel( )
        gm.addModel( PolynomialModel(1) )
        gm.addModel( SineModel() )

        print( gm.shortName( ) )
        print( gm._next.shortName( ) )
        print( gm._next._next.shortName( ) )
        print( gm._next._next._next )

        lolim = numpy.asarray( [-10,-10,  0,-10,-10, 0,-10,-10], dtype=float )
        hilim = numpy.asarray( [ 10, 10, 10, 10, 10, 2, 10, 10], dtype=float )

        gm.setLimits( lolim, hilim )

        ns = NestedSampler( x, gm, y, w )
        ns.verbose = 2
        ns.distribution.setLimits( [0.01, 100] )

        Tools.printclass( ns )

        print( "truth  ", pp )
        self.dofit( ns, pp )

        if plot :
            plotFit( x, y, gm, ftr=ns.samples )

    def nytest( self ) :
        print( "=========== Nested Sampler test 2 ======================" )

        nslim = numpy.asarray( [ 0.01, 100], dtype=float )
        gm.getNoiseScale( ).setLimits(nslim )
        ns = NestedSampler( x, gm, y, w )
        ns.setRate( 0.99 )
        ns.setTrials( 10 )
        ns.setInitialisationEngine( RandomEngine() )
        ns.setEngine( CrossEngine() )
        ns.addEngine( FrogEngine() )
        ns.addEngine( ScaleEngine() )
        ns.addEngine( StepEngine() )
        ns.setVerbose( True )
        np = gm.getNumberOfParameters( )
        ns.setMaxSamples( 1000 )

        print( "truth  ", pp )
        self.dofit( ns, pp )


    def dofit( self, ns, pp ) :
        yfit = ns.sample( )

        par = ns.parameters
        std = ns.standardDeviations
        mlp = ns.samples.maxLikelihoodParameters
        scale = ns.scale
        scdev = ns.stdevScale
        print( "truth  ", fmt( pp ) )
        print( "par    ", fmt( par ) )
        print( "st dev ", fmt( std ) )
        print( "ML par ", fmt( mlp ) )
        print( "scale  ", fmt( scale ), " +- ", fmt( scdev ) )
#        ns.report( )


#        assertAAE( par, mlp, 2 )
#        assertAAE( par, pp, 2 )

    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( NestedSampler1Test.__class__ )





import numpy as numpy
from astropy import units
import math
from . import Tools
from .Formatter import formatter as fmt
from .Engine import Engine

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2017        Do Kester

class StartEngine( Engine ):
    """
    StartEngine generates a random trial sample.

    It is used to initialize the set of trial samples.

    Author       Do Kester.

    """
    #  *********CONSTRUCTORS***************************************************
    def __init__( self, walkers, errdis, copy=None, seed=4213 ):
        """
        Constructor.
        Parameters
        ----------
        copy : StartEngine
            engine to be copied

        """
        super( StartEngine, self ).__init__( walkers, errdis, copy=copy, seed=seed )

    def copy( self ):
        """ Return copy of this.  """
        return StartEngine( self.walkers, self.errdis, copy=self )

    def __str__( self ):
        return str( "StartEngine" )

    #  *********EXECUTE***************************************************
    def execute( self, walker, lowLhood, fitIndex=None ):
        """
        Execute the engine by a random selection of the parameters.

        Parameters
        ----------
        walker : Sample
            sample to diffuse
        lowLhood : float
            lower limit in logLikelihood
        fitIndex : array_like
            list of parameter indices
        Returns
        -------
        int : the number of successfull moves

        """
        if fitIndex is None :
            fitIndex = walker.fitIndex

        model = walker.model
        par = walker.allpars.copy()
        uval = self.rng.rand( len( fitIndex ) )
        par[fitIndex] = self.unit2Domain( model, uval, kpar=fitIndex )
        logL = self.errdis.logLikelihood( model, par )
#        print( "Start  ", walker.id, fmt(uval), fmt(par), fmt(logL) )
        self.setSample( walker, model, par, logL, logW=-math.inf )

        return len( fitIndex )





import math as math

from .Prior import Prior

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2017 Do Kester.

class JeffreysPrior( Prior ):
    """
    Jeffreys prior distribution, for scale-like parameters.
    Jeffreys prior is a improper prior ( i.e. its integral is unbound ).
    Because of that it always needs limits, low and high, such that
    0 < low < high < +Inf.
    .. math::
        Pr( x ) = 1.0 / ( x * norm )   if low < x < high else 0

    where norm = log( high ) - log( low )

    No limits are set by default.

    domain2unit: u = ( log( d ) - log( lo ) ) / ( log( hi ) - log( lo ) );
    unit2domain: d = exp( u * ( log( hi ) - log( lo ) ) + log( lo ) );

    Hidden Attributes
    -----------------
    _logLo : float
        log of lowLimit
    _norm : float
        log of highLimit / lowLimit

    The default of lowLimit and _lowDomain is zero.

    """

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, limits=None, prior=None ):
        """
        Default constructor.

        Parameters
        ----------
        limits : list of 2 floats
            2 limits resp. low and high
        prior : JeffreysPrior
            prior to copy (with new limits if applicable)
        """
        super( JeffreysPrior, self ).__init__( limits=limits, prior=prior )

        #   private attributes at default values
        object.__setattr__( self, "_logLo", -math.inf )
        object.__setattr__( self, "_norm", math.inf )
        object.__setattr__( self, "_lowDomain", 0 )

        if limits is None and prior is not None :
            limits = ( prior.lowLimit, prior.highLimit )
        if limits is not None :
            self.setLimits( limits )
        else :
            self.lowLimit = 0.0

    def copy( self ):
        return JeffreysPrior( prior=self )

    def __setattr__( self, name, value ) :
        if name in ["_logLo", "_norm"] :
            object.__setattr__( self, name, float( value ) )
        else :
            super( JeffreysPrior, self ).__setattr__( name, value )

        if name == "lowLimit" or name == "highLimit" :
            self._logLo = -math.inf if ( self.lowLimit <= 0 ) else math.log( self.lowLimit )
            self._norm = math.log( self.highLimit ) - self._logLo

    def getIntegral( self ) :
        """
        Return the integral of JeffreysPrior from lowLimit to highLimit.
        """
        return self._norm

    def domain2Unit( self, dval ):
        """
        Return a value in [0,1] given a value within the valid domain of
        a parameter for a Jeffreys distribution.

        Parameters
        ----------
        dval : float
            value within the domain of a parameter

        """
        u = ( math.log( dval ) - self._logLo ) / self._norm
        if math.isnan( u ) :
            raise AttributeError( "Limits are needed for JeffreysPrior" )
        return u


    def unit2Domain( self, uval ):
        """
        Return a value within the valid domain of the parameter given a value
        between [0,1] for a Jeffreys distribution.

        Parameters
        ----------
        uval : float
            value within [0,1]

        """
        d = math.exp( uval * self._norm + self._logLo )
        if math.isnan( d ) :
            raise AttributeError( "Limits are needed for JeffreysPrior" )
        return d

    def result( self, x ):
        """
        Return a the result of the distribution function at x.

        Parameters
        ----------
        x : float
            value within the domain of a parameter

        """
        r = 0 if self.isOutOfLimits( x ) else 1.0 / ( x * self._norm )
        if math.isnan( r ) :
            raise AttributeError( "Limits are needed for JeffreysPrior" )
        return r

# logResult has no better definition than the default: just take the math.log of result.
# No specialized method here.

    def partialLog( self, p ):
        """
        Return partial derivative of log( Prior ) wrt parameter.

        Parameters
        ----------
        p : float
            the value

        """
        return math.nan if self.isOutOfLimits( p ) else -1 / p

    def isBound( self ):
        """ Return true if the integral over the prior is bound.  """
        return self.hasLowLimit( ) and self.hasHighLimit( )

    def __str__( self ):
        """ Return a string representation of the prior.  """
        return str( "Jeffreys prior " + ( "unbound." if not self.isBound( )
            else ( "between %.2f and %.2f"%( self.lowLimit, self.highLimit ) ) ) )

#      * End of JeffreysPrior





import numpy as numpy
from . import Tools
from astropy import units
from .LinearModel import LinearModel

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2017 Do Kester

class ChebyshevPolynomialModel( LinearModel ):
    """
    Chebyshev polynomial model of arbitrary degree.

    .. math::
        f( x:p ) = \sum; p_k * T_k( x )

    where the sum is over k running from 0 to degree ( inclusive ).

    The T( x ) are Chebyshev polynomials of the first kind which are defined
    recursively as:
    .. math::
        T_0( x ) = 1\\
        T_1( x ) = x\\
        T_n( x ) = 2 x T_{n-1}( x ) - T_{n-2}( x ) for n >= 2

    These polynomials are orthogonal, only when x is in [-1,+1].

    It is a linear model.

    Examples
    --------
    >>> poly = ChebyshevPolynomialModel( 3 )         # 3rd degree polynomial
    >>> print poly.getNumberOfParameters( )
    4

    """
    def __init__( self, degree, copy=None, **kwargs ):
        """
        Chebyshev Polynomial of a certain degree.
        <br>
        The number of parameters is ( degree + 1 )

        Parameters:  degree the degree of the polynomial.

        """
        names = ["coeff_%d"%k for k in range( degree + 1 )]
        super( ChebyshevPolynomialModel, self ).__init__( degree + 1, copy=copy,
                    names=names, **kwargs )
        object.__setattr__( self, "degree", degree )

    def copy( self ):
        """ Copy method.  """
        return ChebyshevPolynomialModel( self.degree, copy=self )

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the xdata value.

        The partials are calculated using the recurrence formula
        .. math::
            f_n( x ) = 2 * x * f_{n-1}( x ) - f_{n-2}( x )

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the partials
        params : array_like
            parameters of the model (ignored for linear models)
        parlist : array_like
            list of indices of active parameters

        """
        np = self.npmax
        nxdata = Tools.length( xdata )

        part = numpy.zeros( ( nxdata, np ), dtype=float )

        part[:,0] = 1.0
        if np >= 2:
            part[:,1] = xdata
        for i in range( 2, np ):
            part[:,i] = 2 * xdata * part[:,i-1] - part[:,i-2]
        if parlist is None :
            return part

        return part[:,parlist]

    def baseDerivative( self, xdata, params ):
        """
        Returns the derivative df/dx at the xdata value.
        .. math::
            df_n = n * U_{n-1}

        where
        .. math::
            U_0 = 1\\
            U_1 = 2x\\
            U_{n+1} = 2 * x * U_n - U_{n-1}

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the partials
        params : array_like
            parameters of the model

        """
        if self.npmax <= 1 :
            return numpy.zeros_like( xdata )

        np = self.npmax - 1

        um = numpy.ones_like( xdata )
        der = um * params[1]
        if np >= 2:
            uu = 2 * xdata
            der += 2 * uu * params[2]
        for i in range( 2, np ):
            up = 2 * xdata * uu - um
            um = uu
            uu = up
            der += (i+1) * uu * params[i+1]
        return der

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        bn = "ChebyshevPolynomial: f( x:p ) = p_0 * T_0(x)"
        for k in range( 1, self.npmax ):
            bn += " + p_%d * T_%d(x)"%(k,k)
        return bn

    def baseParameterUnit( self, k ):
        """
        Return the unit of the indicated parameter.
        It is always yUnit, as it cannot be otherwise.
        The xUnit must be dimensionless.

        Parameters
        ----------
        k : int
            parameter number.

        """
        return self.yUnit





import numpy as numpy
from astropy import units
from scipy.optimize import curve_fit
import math
from . import Tools
from .Formatter import formatter as fmt

from .ConvergenceError import ConvergenceError
from .BaseFitter import BaseFitter
from .IterativeFitter import IterativeFitter
from .IterationPlotter import IterationPlotter

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2017 Do Kester

class CurveFitter( IterativeFitter ):
    """
    CurveFitter implements scipy.optimize.curve_fit.

    Author:      Do Kester.

    Attributes
    ----------
    method : {'lm', 'trf', 'dogbox'}
        'lm'        LevenbergMarquardt (default for no limits)
        'trf'       Trust Region Reflective (default for limits)
        'dogbox'    for small problems with limits

    Raises
    ------
    ConvergenceError    Something went wrong during the convergence if the fit.

    """
    def __init__( self, xdata, model, method=None, fixedScale=None, map=False, keep=None ):
        """
        Create a new class, providing inputs and model.

        Parameters
        ----------
        xdata : array_like
            array of independent input values
        model : Model
            the model function to be fitted
        method : 'lm' | 'trf' | 'dogbox'
            method to be used
        fixedScale : float
            the fixed noise scale.
        map : bool (False)
            When true, the xdata should be interpreted as a map.
            The fitting is done on the pixel indices of the map,
            using ImageAssistant
        keep : dict of {int:float}
            dictionary of indices (int) to be kept at a fixed value (float)
            The values of keep will be used by the Fitter as long as the Fitter exists.
            See also `fit( ..., keep=dict )`

        """
        super( CurveFitter, self ).__init__( xdata, model, map=map, keep=keep,
                            fixedScale=fixedScale )
        self.method = method

    #  *************************************************************************
    def fit( self, ydata, weights=None, inipar=None, keep=None, limits=None,
            plot=False, **kwargs ):
        """
        Return      parameters for the model fitted to the data array.

        Parameters
        ----------
        ydata : array_like
            the data vector to be fitted
        weights : array_like
            weights pertaining to the data
            The weights are relative weights unless fixedScale is set.
        inipar : array_like
            inital parameters (default from Model)
        keep :  dict of {int:float}
            dictionary of indices (int) to be kept at a fixed value (float)
            The values of keep are only valid for *this* fit
            See also `CurveFitter( ..., keep=dict )`
        limits : None or list of 2 floats or list of 2 array_like
            None :        from Model if Model has limits set else no limits
            [-inf,+inf] : no limits applied
            [lo,hi] :     low and high limits for all values
            [la,ha] :     low array and high array limits for the values
        plot : bool
            Plot the results.
        kwargs : dict
            keywords arguments to be passed to :ref:`curve_fit<scipy.optimize.curve_fit>`

        Raises
        ------
            ValueError when ydata or weights contain a NaN
        """
        fitIndex = self.fitprolog( ydata, weights=weights, keep=keep )

        abssigma = ( self.fixedScale is not None )
        if weights is None :
            sigma = ( None if not abssigma else
                      numpy.full_like( ydata, 1 / math.sqrt( self.fixedScale ), dtype=float ) )
        else :
            wgts = numpy.where( weights == 0, 1e-20, weights )      # avoid division by 0
            sigma = 1 / numpy.sqrt( wgts )

        # Check for limits
        if limits is not None :
            bounds = limits
        elif self.model.priors :
            bounds = ( self.model.lowLimits, self.model.highLimits )
        else :
            bounds = ( -numpy.inf, numpy.inf )

        # Make requested fitIndex; save original one.
        if keep is not None :
            self.fitIndex = self.keepFixed( keep )

        # Make initial parameters
        p0 = self.model.parameters if inipar is None else inipar
        if fitIndex is not None :
            p0 = p0[fitIndex]

        ydat = ydata
        # Normalize the data when needed.
        if hasattr( self, "normdata" ) :
            ydat = numpy.append( ydata, self.normdata )
            if sigma is None :
                sigma = numpy.ones_like( ydata, dtype=float )
            sigma = numpy.append( sigma, 1.0 / numpy.sqrt( self.normweight ) )

        # Call scipy.curve_fit
        out = curve_fit( self.result, self.xdata, ydat, p0=p0,
            sigma=sigma, absolute_sigma=abssigma, bounds=bounds, jac=self.jacobian,
            method=self.method, **kwargs )

        pars = self.insertParameters( out[0], index=fitIndex )
        pcov = out[1]
        self.model.parameters = pars
        self.covariance = pcov
        self.chiSquared( ydata, weights=weights )

        self.fulloutput = out

        self.fitpostscript( ydata, plot=plot )

        return pars

    def __str__( self ) :
        return "CurveFitter"


# ================================================================================
# The following methods `result` and `jacobian` are passed to scipy's curvefit.
#
# Both have a 'starred' fitpar: (*fitpar) as curvefit expect a tuple of parameters.
# We prefer fitpar to be an array_like thing.
# ================================================================================

    def result( self, xdata, *fitpar ) :
        """
        Result method to make connection to the scipy optimizers

        Parameters
        ----------
        xdata : array_like
            input data
        fitpar : tuple of float
            parameters for the model
        """
        params = Tools.toArray( fitpar, dtype=float, ndim=1 )
        params = self.insertParameters( params )
        res = self.model.result( xdata, params )
        if hasattr( self, "normdfdp" ) :
            extra = numpy.inner( self.normdfdp, params )
            res = numpy.append( res, [extra] )

        return res

    def jacobian( self, xdata, *fitpar ) :
        """
        Method to make connection to the scipy optimizers

        Parameters
        ----------
        xdata : array_like
            input data
        fitpar : (tuple of) float
            parameters for the model
        """
        params = Tools.toArray( fitpar, dtype=float, ndim=1 )
        params = self.insertParameters( params )

        design = self.getDesign( params=params, xdata=xdata, index=self.fitIndex )
        return design




import numpy as numpy
import random
from astropy import units
from . import Tools
import warnings
from .Formatter import formatter as fmt

from .Prior import Prior

__author__ = "Do Kester"
__year__ = 2018
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2016 - 2018 Do Kester


class Dynamic( object ):
    """
    Class adjoint to Model which implements some dynamic behaviour.

    """

    def getNumberOfComponents( self ):
        return self.npbase / self.deltaNpar

    def __setattr__( self, name, value ) :
        dind = {"growPrior" : Prior}
        if ( Tools.setNoneAttributes( self, name, value, lnon ) or
             Tools.setSingleAttributes( self, name, value, dind ) ) :
            pass
        else :
            super( ).__setattr__( name, value )






    def alterParameterSize( self, dnp, pat ) :
        """
        Change the number of parameters and self.parameters.

        Parameters
        ----------
        npnew : int
            number of parameters in the DynamicModel
        pat : int
            starting index of the DynamicModel
        """
        npnew = self.npbase + dnp
#        print( "Dyn  ", pat, self.npbase, dnp )

        if dnp == 0 :
            return

        mdlpar = self._head.parameters
#        print( "mdlpar  ", mdlpar )
        mdlpar = self.alterParameters( mdlpar, self.npbase, dnp, pat )

        self.npmax += dnp
        self.npbase += dnp
        self._head._npchain += dnp

        self._head.parameters = mdlpar

    def alterParameters( self, param, npbase, dnp, pat ) :

        newpar = numpy.zeros( len( param ) + dnp, dtype=float )
        kh = pat + npbase + dnp
        kb = pat + npbase

#        print( param.shape, npbase, pat, dnp, kh, kb )
        if dnp < 0 :
            newpar[:kh] = param[:kh]
            newpar[kh:] = param[kb:]
        else :
            newpar[:kb] = param[:kb]
            newpar[kh:] = param[kb:]

#        print( "nwpar  ", fmt( newpar, max=None ) )
        return newpar

    def alterFitindex( self, findex, npbase, dnp, pat ) :

        newfi = numpy.zeros( len( findex ) + dnp, dtype=int )
        kh = pat + npbase + dnp
        kb = pat + npbase

#        print( param.shape, npbase, pat, dnp, kh, kb )
        if dnp < 0 :
            newfi[:kh] = findex[:kh]
            newfi[kh:] = findex[kb:]
        else :
            newfi[:kb] = findex[:kb]
            newfi[kb:kh] = numpy.arange( dnp ) + findex[kb-1] + 1
            newfi[kh:] = findex[kb:]

#        print( "newfi  ", fmt( newfi, max=None ) )
        return newfi







import numpy as numpy
import math

from .Kernel import Kernel

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  *    This file is part of the BayesicFitting package.
#  *
#  *    BayesicFitting is free software: you can redistribute it and/or modify
#  *    it under the terms of the GNU Lesser General Public License as
#  *    published by the Free Software Foundation, either version 3 of
#  *    the License, or ( at your option ) any later version.
#  *
#  *    BayesicFitting is distributed in the hope that it will be useful,
#  *    but WITHOUT ANY WARRANTY; without even the implied warranty of
#  *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  *    GNU Lesser General Public License for more details.
#  *
#  *    The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2013 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2017 Do Kester

class Parabola( Kernel ):
    """
    Parabola is a Kernel function between [-1,1]; it is 0 elsewhere.
    .. math::
        K( x ) = ( 1-x^2 ) if |x| < 1 else 0


    """
    def __init__( self ) :
        super( Parabola, self ).__init__( integral=4.0/3.0, fwhm=2 * math.sqrt(0.5) )

    def result( self, x ):
        return self.resultsq( x * x )

    def resultsq( self, xsq ):
        return numpy.where( xsq <= 1, 1.0 - xsq, 0.0 )

    def partial( self, x ):
        return numpy.where( numpy.abs( x ) <= 1, -2 * x, 0.0 )

    def isBound( self ):
        return True

    def name( self ):
        return str( "Parabola: ( 1-x^2 ) if |x| < 1 else 0" )





import numpy as numpy
from astropy import units
import math
from . import Tools

from .Engine import Engine

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2017        Do Kester

class RandomEngine( Engine ):
    """
    RandomEngine generates a random trial sample.

    It is used to initialize the set of trial samples.

    Author       Do Kester.

    """
    #  *********CONSTRUCTORS***************************************************
    def __init__( self, walkers, errdis, copy=None, seed=4213 ):
        """
        Constructor.
        Parameters
        ----------
        walkers : SampleList
            walkers to be diffused
        errdis : ErrorDistribution
            error distribution to be used
        copy : RandomEngine
            engine to be copied
        seed : int
            for random number generator
        """
        super( RandomEngine, self ).__init__( walkers, errdis, copy=copy, seed=seed )

    def copy( self ):
        """ Return copy of this.  """
        return RandomEngine( self.walkers, self.errdis, copy=self )

    def __str__( self ):
        return str( "RandomEngine" )

    #  *********EXECUTE***************************************************
    def execute( self, walker, lowLhood, fitIndex=None ):
        """
        Execute the engine by a random selection of the parameters.

        Parameters
        ----------
        walker : Sample
            sample to diffuse
        lowLhood : float
            lower limit in logLikelihood
        fitIndex : array_like
            list of parameter indices
        Returns
        -------
        int : the number of successfull moves

        """

        if fitIndex is None :
            fitIndex = walker.fitIndex

        perm = self.rng.permutation( fitIndex )

        model = walker.model

        mlen = len( self.walkers )
        ur = self.unitRange * mlen / ( mlen - 1 )
        um = self.unitMin - ur / mlen
        q = numpy.where( um < 0 )
        um[q] = 0
        ux = um + ur
        q = numpy.where( ux > 1 )
        ux[q] = 1 - um[q]

        t = 0
        for c in perm :
            param = walker.allpars.copy( )
            save = param[c]
            kk = 0
            while True :
                kk += 1
                uval = self.rng.uniform( um[c], ux[c], 1 )
                param[c] = self.unit2Domain( model, uval, c )

                Ltry = self.errdis.updateLogL( model, param, parval={c : save} )
                if Ltry >= lowLhood:
                    self.reportSuccess( )
                    self.setSample( walker, model, param, Ltry )
                    t += 1
                    break
                elif kk < self.maxtrials :
                    self.reportReject( )
                else :
                    self.reportFailed()
                    param[c] = save
                    break

        return t                        # nr of succesfull steps





from .KernelModel import KernelModel
from .kernels.Sinc import Sinc

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2017 Do Kester

class SincModel( KernelModel ):
    """
    Sinc Model.
    Also known as Cardinal Sine.
    .. math::
        f( x:p ) = p_0 * sin( ( x - p_1 ) / p_2 ) / ( ( x - p_1 ) / p_2 )

    where :math:`p_0` = amplitude, :math:`p_1` = offset and
    :math:`p_2` = width ( =Distance between first zero-crossings divided by 2 Pi. )
    As always x = input.

    The parameters are initialized at {1.0, 0.0, 1.0}.
    Parameter 2 ( width ) is always kept positive ( >=0 ).

    SincModel() is syntactic sugar for KernelModel( kernel=Sinc() )

    Examples
    --------
    >>> sinc = SincModel( )
    >>> print sinc.npchain
    3
    >>> print( sinc( numpy.arange( 15, dtype=float )-7 ) )        # sinc function between [-7,+7]
    [0,0,0,0,0,0,0,1,0,0,0,0,0,0,0]

    """
    def __init__( self, copy=None, **kwargs ):
        """
        Sinc model.

        Number of parameters is 3.

        Parameters
        ----------
        copy : ArctanModel
            to be copied
        fixed : dictionary of {int:float}
            int     list if parameters to fix permanently. Default None.
            float   list of values for the fixed parameters.
            Attribute fixed can only be set in the constructor.
        """
        super( SincModel, self ).__init__( kernel=Sinc(), copy=copy, **kwargs )

    def copy( self ):
        """ Copy method.  """
        return SincModel( copy=self )





import numpy as numpy
import math as math

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"


#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2017 Do Kester

class Prior( object ):
    """
    Base class defining prior distributions.
    Most of the lower and upper limit handling is done in this class.

    Two methods need to be defined which map the values between [0,1]
    on to the domain, and vice versa: unit2Domain and domain2Unit.

    The copy method is also necessary.

    Attributes
    ----------
    lowLimit : float
        low limit on the Prior
    highLimit : float
        high limit on the Prior
    deltaP : float
        width of numerical partial derivative calculation

    Hidden Attributes
    -----------------
    _lowDomain : float
        lower limit of the Priors possible values
    _highDomain : float
        upper limit of the Priors possible values

    """

    #*********CONSTRUCTORS***************************************************
    def __init__( self, limits=None, prior=None ):
        """
        Default constructor.

        Parameters
        ----------
        limits : list of 2 floats
            2 limits resp. low and high
        prior : Prior
            prior to copy (with new limits if applicable)
        """
        super( object, self ).__init__()

        self.deltaP = 0.0001
        self._lowDomain = -math.inf
        self._highDomain = math.inf

        object.__setattr__( self, "deltaP", 0.001 )             # for numerical partials
        object.__setattr__( self, "lowLimit", -math.inf )       # Lower limit.
        object.__setattr__( self, "highLimit", math.inf )       # Upper limit.
#       private properties of the Prior
        object.__setattr__( self, "_lowDomain", -math.inf )     # Lower limit of the Priors possible domain.
        object.__setattr__( self, "_highDomain", math.inf )     # Upper limit of the Priors possible domain

        self.setLimits( limits )
        if prior is not None :
            self.deltaP = prior.deltaP

    def copy( self ) :
        """ Return a copy """
        return Prior( prior=self, limits=self.limits )

    def setLimits( self, limits ):
        """
        Set limits.
        It is asserted that limits[0] is smaller than limits[1].

        Parameters
        ----------
        limits : None or list of 2 float
            None : limits are set to [-inf,+inf]
            the [low,high] limits.
        Raises
        ------
        ValueError when low limit is larger than high limit

        """
        if limits is None :
            limits = [-math.inf,+math.inf]

        if limits[0] > limits[1] :
            raise ValueError( "The lowlimit %f must be smaller than the high limit %f"%
                                ( limits[0], limits[1] ) )
        if not ( self._lowDomain <= limits[0] < limits[1] <= self._highDomain ) :
            raise ValueError( "Limits out of order or out of domain" )

        self.lowLimit = limits[0]
        self.highLimit = limits[1]

    def __setattr__( self, name, value ) :
        """
        Set attributes: lowLimit, highLimit, deltaP, _lowDomain, _highDomain.

        Also set scale for Priors that need a scale.

        """
        keys = ["lowLimit", "highLimit", "deltaP", "scale", "_lowDomain", "_highDomain"]
        if name == "scale" and value <= 0 :
            raise ValueError( "Scale must be positive" )

        if name in keys :
            object.__setattr__( self, name, float( value ) )
        else :
            raise AttributeError( repr( self ) + " object has no attribute " + name )

    def unsetLimits( self ):
        """ Remove all limits.  """
        self.lowLimit = self._lowDomain
        self.highLimit = self._highDomain

    def isOutOfLimits( self, par ):
        """
        True if the parameter is out of limits

        Parameters
        ----------
        par : float
            the parameter to check

        """
        return ( par < self.lowLimit ) or ( par > self.highLimit )

    def checkLimit( self, par ):
        """
        Check whether the parameter is within limits.

        Parameters
        ----------
        par : float
            the parameter to check

        Raises
        ------
            ValueError when outside limits.

        """
        if self.isOutOfLimits( par ):
            raise ValueError( "Parameter outside supplied limits" )

    def stayInLimits( self, par ):
        """
        Return lower limit or upper limit when parameter is outside.

        Parameters
        ----------
        par : float
            the parameter to check

        """
        if par < self.lowLimit:
            return self.lowLimit
        if par > self.highLimit:
            return self.highLimit
        return par

    def hasLowLimit( self ):
        """ Return true if the prior has its low limits set.  """
        return self.lowLimit > self._lowDomain

    def hasHighLimit( self ):
        """ Return true if the prior has its high limits set.  """
        return self.highLimit < self._highDomain

    def hasLimits( self ):
        """ Return true if it has any limits.  """
        return self.hasLowLimit() and self.hasHighLimit()

    def getLimits( self ):
        """ Return the limits.  """
        return numpy.array( [ self.lowLimit, self.highLimit ] )

    def getIntegral( self ) :
        """
        Return the integral of the prior over the valid range.

        Default: 1.0 (for bound priors)
        """
        return 1.0

    def getRange( self ):
        """ Return the range.  """
        return self.highLimit - self.lowLimit

    def domain2Unit( self, dval ):
        """
        Return a value in [0,1] given a value within the valid domain of
        a parameter for a distribution.

        Parameters
        ----------
        dval : float
            value within the domain of a parameter

        """
        pass

    def unit2Domain( self, uval ):
        """
        Return a value within the valid domain of the parameter given a value
        between [0,1] for a distribution.

        Parameters
        ----------
        uval : float
            value within [0,1]

        """
        pass

    def result( self, p ):
        """
        Return value of the Prior at a given value.

        If result is not defined, fall back to numerical derivative od Domain2Unit.

        Parameters
        ----------
        p : float
            the value

        """
        return self.numPartialDomain2Unit( p )
#        return 0.0 if self.isOutOfLimits( p ) else self.partialDomain2Unit( p )

    def partialDomain2Unit( self, p ):
        """
        Return the derivative of Domain2Unit, aka the result of the distribution at p

        Parameters
        ----------
        p : float
            the value

        """
        return self.result( p )

    def logResult( self, p ) :
        """
        Return the log of the result; -inf when p == 0.

        Parameters
        ----------
        p : float
            the value

        """
        try :
            return math.log( self.result( p ) )
        except :
            return -math.inf

#    def partialDomain2Unit( self, dval ):
        """
        Return a the derivate of the domain2Unit function to dval.
        By default: numeric derivative.

        Parameters
        ----------
        dval : float
            value within the domain of a parameter

        """
#        return self.numPartialDomain2Unit( dval )

    def numPartialDomain2Unit( self, dval ):
        """
        Return a the numeric derivate of the domain2Unit function to dval.

        Parameters
        ----------
        dval : float
            value within the domain of a parameter

        """
        return ( ( self.domain2Unit( dval + self.deltaP ) -
                   self.domain2Unit( dval - self.deltaP) ) /
                        ( 2 * self.deltaP ) )

    def partialLog( self, p ):
        """
        Return partial derivative of log( Prior ) wrt parameter.
        default numPartialLog

        Parameters
        ----------
        p : float
            the value

        """
        return self.numPartialLog( p )

    def numPartialLog( self, p ):
        """
        Return the numeric partial derivative of log( Prior ) wrt parameter.
        Parameters
        ----------
        p : float
            the value

        """
        if self.isOutOfLimits( p ) : return math.nan
#        if math.isinf( p ) : return -p
        rm = self.logResult( p - self.deltaP )
        rp = self.logResult( p + self.deltaP )
        return ( rp - rm ) / ( 2 * self.deltaP )

    def isBound( self ):
        """ Return true if the integral over the prior is bound.  """
        return False

    def __str__( self ):
        """ Return a string representation of the prior.  """
        pass

#      * End of Prior





# run with : python3 -m unittest TestFixedModels.TestFixedModels.testGaussModel

import unittest
import numpy as numpy
from astropy import units
import matplotlib.pyplot as plt
import warnings

from StdTests import stdModeltest
from BayesicFitting import *


__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class TestFixedModels( unittest.TestCase ):
    """
    Test harness for Models

    Author:      Do Kester

    """
    def plotArctanModel( self ) :
        self.testArctanModel( plot=True )

    def plotConstantModel( self ) :
        self.testConstantModel( plot=True )

    def plotEtalonModel( self ) :
        self.testEtalonModel( plot=True )

    def plotExpModel( self ) :
        self.testExpModel( plot=True )

    def plotVoigtModel( self ) :
        self.testVoigtModel( )
        xx = numpy.linspace( -1, +1, 1001 )
        par = numpy.asarray( [1.2,-0.2,0.3,0.3], dtype=float )
        plt.plot( xx, VoigtModel().result( xx, par ), '-', linewidth=2 )
        plt.plot( xx, GaussModel().result( xx, par[[0,1,2]] ), 'r-' )
        pg = par.copy()
        pg[3] = 0
        plt.plot( xx, VoigtModel().result( xx, pg ), 'r--' )
        plt.plot( xx, LorentzModel().result( xx, par[[0,1,3]] ), 'g-' )
        pg = par.copy()
        pg[2] = 0
        plt.plot( xx, VoigtModel().result( xx, pg ), 'g--' )
        plt.show()

    def plotLorentzModel( self ) :
        self.testLorentzModel( plot=True )

    def plotFreeShapeModel( self ) :
        self.XtestFreeShapeModel( plot=True )

    def plotGaussModel( self ) :
        self.testGaussModel( plot=True )

    def plotGaussPlusBackgroundModel( self ) :
        self.testGaussPlusBackgroundModel( plot=True )

    def plotHarmonicModel( self ) :
        self.testHarmonicModel( plot=True )

    def plotPolynomialModel( self ) :
        self.testPolynomialModel( plot=True )

    def plotChebyshevPolynomialModel( self ) :
        self.testChebyshevPolynomialModel( plot=True )

    def plotPadeModel( self ) :
        self.testPadeModel( plot=True )

    def plotPowerLawModel( self ) :
        self.testPowerLawModel( plot=True )

    def plotPowerModel( self ) :
        self.testPowerModel( plot=True )

    def plotSincModel( self ) :
        self.testSincModel( plot=True )

    def plotSineModel( self ) :
        self.testSineModel( plot=True )

    def plotSineAmpModel( self ) :
        self.testSineAmpModel( plot=True )

    def plotSineSplineModel( self ) :
        self.testSineSplineModel( plot=True )

    def plotSineSplineDriftModel( self ) :
        self.testSineSplineDriftModel( plot=True )

    def plotSineDriftModel( self ) :
        self.testSineDriftModel( plot=True )

    def plotSplinesModel( self ) :
        self.testSplinesModel( plot=True )


    def testArctanModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******ARCTAN******************" )
        m = ArctanModel( fixed={1:-1.2} )
        p = numpy.asarray( [1.2,30], dtype=float )

        stdModeltest( m, p, plot=plot )

    def testEtalonModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******ETALON***********************" )
        m = EtalonModel( fixed={0:1.5})
        p = numpy.asarray( [30.0, 2.0, 0.2], dtype=float )

        stdModeltest( m, p, plot=plot )

    def testEtalonModel2( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******ETALON 2***********************" )
        fm = PolynomialModel( 1 )
        m = EtalonModel( fixed={0:1.5, 1:fm})
        p = numpy.asarray( [2.0, 0.2, 1.0, 0.9], dtype=float )

        stdModeltest( m, p, plot=plot )

    def testEtalonModel3( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******ETALON 2***********************" )
        fm = PolynomialModel( 1 )
        fm += SineModel()
        am = SplinesModel( nrknots=2, min=-1, max=1 )
        m = EtalonModel( fixed={0:am, 1:fm})
        par = [10.0, 0.2, 1.0, 0.1, 0.0, 0.02, 1.0, 0.5, 2.0, 0.1, 0.0]
        p = numpy.asarray( par, dtype=float )

        stdModeltest( m, p, plot=plot )

    def testExpModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******EXP**************************" )
        m = ExpModel( fixed={1:0.1} )
        p = numpy.asarray( [1.2], dtype=float )

        stdModeltest( m, p, plot=plot )

    def testLorentzModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******LORENTZ***********************" )
        m = LorentzModel( fixed={0:4} )
        p = numpy.asarray( [-0.2,0.3], dtype=float )

        stdModeltest( m, p, plot=plot )

    def testVoigtModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******VOIGT***********************" )
        m = VoigtModel( fixed={1:0.2,2:0.3} )
        p = numpy.asarray( [1.2,0.4], dtype=float )
        print( m.parameters )
        print( m.expand( x, m.parameters ) )

        stdModeltest( m, p, plot=plot, warn=["nopart"] )

    def XtestFreeShapeModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******FREESHAPE********************" )

        self.assertRaises( AttributeError, FreeShapeModel,  5, pixperbin=1, xlo=-1,
                    xhi=1.1, fixed={3:1.1} )

    def testGaussModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******GAUSS*************************" )
        m = GaussModel( fixed={2:0.3} )
        p = numpy.asarray( [1.2,-0.2], dtype=float )

        stdModeltest( m, p, plot=plot )

    def testGaussPlusBackgroundModel( self, plot=False ):
        print( "******GAUSS + BG**********************" )
        gm = GaussModel( fixed={1:0.1} )
        print( gm )
        print( gm.parameters )
        pm = PolynomialModel( 2, fixed={0:2.9} )
        print( pm )
        print( pm.parameters )
        gm.addModel( pm )
        self.assertTrue( gm.npchain == 4 )
        par = numpy.asarray( [3,0.2,0.1,0.1], dtype=float )

        stdModeltest( gm, par, plot=plot )

    def testHarmonicModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******HARMONIC**********************" )
        m = HarmonicModel( order=3, period=0.4, fixed={0:1.0,2:1.0,4:1.0} )
        self.assertTrue( m.order == 3 )
        self.assertTrue( m.npbase == 3 )
        self.assertTrue( m.npmax == 6 )
        self.assertTrue( m.period == 0.4 )

        p = numpy.asarray( [1.2,0.3,1.2], dtype=float )

        stdModeltest( m, p, plot=plot )

    def testPolynomialModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******POLYNOMIAL*********************" )
        sm = SineModel()
        m = PolynomialModel( 2, fixed={0:sm} )
        Tools.printclass( m )
        self.assertTrue( m.getNumberOfParameters( ) == 5 )
        self.assertTrue( m.npbase == 5 )
        self.assertTrue( m.npmax == 3 )
        p = numpy.asarray( [0.2,0.1,1.9,-1.2,1.3], dtype=float )
        print( sm.partial( x, p[2:] ) )
        print( m.partial( x, p ) )
        pm = PolynomialModel( 2, fixed={0:1.0} )
        pm *= SineModel()
        print( pm )
        print( pm.result( x, p ) )
        print( m )
        print( m.result( x, p ) )
        stdModeltest( m, p, plot=plot )

    def testChebyshevPolynomialModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******CHEBYSHEV POLYNOMIAL*************" )
        m = ChebyshevPolynomialModel( 4, fixed={4:0.1} )
        self.assertTrue( m.getNumberOfParameters( ) == 4 )
        self.assertTrue( m.npbase == 4 )
        p = numpy.asarray( [1,-2,3,-2], dtype=float )
        stdModeltest( m, p, plot=plot )

    def testPadeModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******PADE***********************" )
        m = PadeModel( 3, 2, fixed={0:1} )
        self.assertTrue( m.getNumberOfParameters( ) == 6 )
        self.assertTrue( m.npbase == 6 )
        self.assertTrue( m.npmax == 7 )
        p =  numpy.asarray( [-2, 3, -2, 0.3, 1, 0.2], dtype=float )
        stdModeltest( m, p, plot=plot )

    def testPowerLawModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******POWERLAW*******************" )
        m = PowerLawModel( fixed={1:-2} )
        self.assertTrue( m.npchain == 2 )
        self.assertTrue( m.npbase == 2 )
        p = [2.3, 0.5]
        stdModeltest( m, p, plot=plot )

    def testPowerModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******POWER**********************" )
        m = PowerModel( 3, fixed={0:1.2} )
        self.assertTrue( m.getNumberOfParameters( ) == 0 )
        self.assertTrue( m.npbase == 0 )
        p = []
        stdModeltest( m, p, plot=plot )

    def testSincModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******SINC***********************" )
        m = SincModel( fixed={2:0.1,1:2} )
        self.assertTrue( m.npchain == 1 )
        self.assertTrue( m.npbase == 1 )
        p = [2.3]
        stdModeltest( m, p, plot=plot )

    def testSineModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******SINE***********************" )
        m = SineModel( fixed={0:1})
        self.assertTrue( m.npchain == 2 )
        self.assertTrue( m.npbase == 2 )
        p = [-1.1, 0.5]
        stdModeltest( m, p, plot=plot )

    def testSineAmpModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******SINEAMP********************" )
        m = SineAmpModel( 1.3, fixed={1:1} )
        self.assertTrue( m.npchain == 1 )
        self.assertTrue( m.npbase == 1 )
        p = [-1.1]
        stdModeltest( m, p, plot=plot )

    def testSineDriftModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******SINEDRIFT******************" )

        m = SineDriftModel( fixed={2:0.8} )
        self.assertTrue( m.npmax == 4 )
        self.assertTrue( m.npchain == 3 )
        self.assertTrue( m.npbase == 3 )
        p = [1.1, 0.5, 0.8, 0.4 ]
        stdModeltest( m, p, plot=plot )

    def testSplinesModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******SPLINES*************************" )
        knots = numpy.arange( 11, dtype=float )
        self.assertRaises( AttributeError, SplinesModel, knots, fixed={3:1.1} )

    def testSineSplineModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******SINESPLINE******************" )
        knots = numpy.arange( 3, dtype=float ) - 1
        freq = 150.0
        self.assertRaises( AttributeError, SineSplineModel, freq, knots, fixed={3:1.1} )


    def testSineSplineDriftModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******SINESPLINEDRIFT*************" )
        knots = numpy.arange( 3, dtype=float ) - 1
        self.assertRaises( AttributeError, SineSplineDriftModel, knots, fixed={3:1.1} )




    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestModels.__class__ )

if __name__ == '__main__':
    unittest.main( )


